

/* 
	Rare, odd, and proprietary file formats: Import functions
*/


/*		MODULE FOR CRYSTAL IMAGE 0.85 BETA
		This program is copyrighted material.
		(C) 2002-2015 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>



#include "includeall.h"









/************************************************************************************

	Import wim files from the AstroPhysics X-ray line scanner.
	It starts with a 264 byte header where the last 12 bytes are relevant.
	The wim format is an byte-interlaced format for low and high energy
	respectively.

*************************************************************************************/



int  wim_read (image_type* img, const char* filename)
{

FILE* FP;
unsigned char header[264];
int width, height, numslices, darkval, magic;
long numelems,i,j;
unsigned char *p, *strip;
int x,y;


	FP = fopen (filename, "rb");
	if (!FP) return -1;

	numelems = fread (header, sizeof (char), 264, FP);	/* read in header info */
	if (numelems != 264)
	{
		dp (0, "Error: WIM header: Read %d elements, should be 264\n", numelems);
		return -1;
	}
	magic = header[252]+ ((header[253])<<8);
	darkval = header[258] + ((header[259])<<8);
	height = header[260] + ((header[261])<<8);
	width = header[262] + ((header[263])<<8);
	numslices = 2;
	for (i=0; i<264; i++)
	{
		dp (2,"%5d   %3d  \n", i,header[i]);
	}
	
	dp (1, "WIM file header: Magic %d. Size %dx%d, dark value %d\n",magic, width,height,darkval);
	if (magic != 0x020C) dp (0, "Warning: Possibly not a WIM file\n");
	
	/* We found the image size, now read strip by strip and sort into img->data */
	
	img->xmax = width;
	img->ymax = height; 
	img->zmax = numslices;
	img->maxval.imax=255;
	img->minval.imin=0;
	img->imgtype=B_CHAR;
	x = width;
	y = height;
	
	img->data = calloc (img->xmax*img->ymax*img->zmax,1);
	if (!img->data) 
	{
		fclose (FP);
		fatal_memerr();
	}
	p = img->data;

	strip = calloc (2*x+256, sizeof (unsigned char));
	if (!strip) fatal_memerr();

	
	/* Each strip contains pairs of bytes:
	   first byte, later slice 0, is high energy;
	   second byte, later lice 1, is low energy.
	*/
	
	for (i=0; i<y; i++)
	{
		numelems = fread (strip, sizeof (unsigned char), 2*x, FP);
		for (j=0; j<x; j++)
		{
			p[i*x+j] = strip[2*j];
			p[i*x+j+x*y] = strip[2*j+1];
		}
	}

	fclose (FP);
	free (strip);
	return 0;
	
}




/************************************************************************************

	Import Compix SimplePCI (.cxd) files
	The cxd format is pretty complex, and here, we implement only a rudimentary
	reader for the most common confocal formats.
	For a more advanced importer, check out PCIreader.java published by the
	Open Microscopy Environment
	For the POIFS, see http://poi.apache.org/poifs/fileformat.html
	Even more comprehensive (though less comprehensible) information can be obtained 
	from M$ here: https://msdn.microsoft.com/en-us/library/dd942138.aspx

	Oh, and can somebody at the Apache Foundation please explain what POI stands for?

*************************************************************************************/

/* Some info about the blocks:

	1) HEADER. The first 512 bytes constitute the header, which contains
		general file info (see struct poifs_type)

	2) Block Allocation Table (BAT)
		Consists of multiple sectors. The first BAT sector is block 0 (0x200 to 0x3FF),
		but consecutive BAT sectors are found indexed in the header's INTs at 0x004C, 
		0x0050, 0x0054, ... 0x01FC. Each 512-byte sector pointed at by one of these INTs 
		holds the BAT entry for 128 blocks. Thus, the BAT can address 109 x 128 sectors
		(13952 sectors = 6.8 MB).

		The BAT contains integers that point to the next block in a chain. If we interpret
		the BAT as an array int bat[13952], then bat[0] belongs to block 0, bat[1] belongs to
		block 1 etc. An object (file) contains a start block, say, block A. We read block A
		into the data buffer, but we also read bat[A] to obtain B, where B is the next block.
		Thus, we append block B and read bat[B] for the next block etc until bat[n] results in
		-2 (end of chain)

		Additional codes are:
			-1 (0x FF FF FF FF)		Unallocated sector
			-2 (0x FF FF FF FE)		End of chain, see above
			-3 (0x FF FF FF FD)		The corresponding sector is a BAT sector
			-4 (0x FF FF FF FC)		The corresponding sector is an XBAT sector (??)
		Note that as per Microsoft, these are unsigned INTs, so a value of 0xFFFFFFF9 is 
		a positive, valid sector number.

	3) Extended Block Allocation Table (XBAT) Index
		The XBAT index simply continues the BAT index. Where the BAT index can hold 109 BAT
		blocks, the XBAT holds the 110th and subsequent BAT block indices. So we obtain the
		sequence of BAT blocks 0 to 108 from the header and subsequent ones from the XBAT.
		The XBAT is read as a linked chain where the last entry (integer #127) points
		at the next block or holds -2 for the end of chain.

	4) Short Block Allocation Table (SBAT) and index
		The mini-stream and its associated SBAT are used for files of less than 4096 bytes.
		The idea is the same as for the BAT: Each ministream sector (64 bytes long) has an
		associated SBAT entry that chains the next ministream block. The SBAT can be read
		much like a regular data stream via the BAT. The ministream needs to be read via BAT,
		too, but eight ministream sectors fit into one regular sector. Thus, the ministream
		has one more level of indirection as we need to translate the ministream block into
		a BAT block and then access that block via regular BAT/XBAT mechanisms.

	5) Properties (can be seen as table of contents)
		Properties contain either the root entry (5), a directory (1) or a file (2)
		The block of the root entry is found in the header, and all properties can be read
		in sequence through its BAT chain. Properties form a tree (next, prev, child),
		in fact, a binary tree (with node colors red and black). 
		Properties have a name, size, and (if applicable) start block for the BAT chain.
		In the practice of cxd files, the tree is pretty useless, and it is preferable to
		just scan the properties for tree snippets.

*/




typedef struct
{
	unsigned char name[34];						/* Name reduced from 16-bit Unicode to 8-bit ASCII */
	unsigned char prop_type;					/* File, Directory, or Root */
	int prev_prop, next_prop, child_prop;		/* For a linked tree, but poorly implemented by Compix */
	int state_flags;							/* Apparently unused */
	int start_blk;								/* Start block in BAT or SBAT stream */
	int size;									/* File size in bytes */
}
poi_prop_type;



typedef struct
{
	int blksize, sblksize;			/* Size of one block (and one small block), usually 512 and 64 */
	int bat_list[109];				/* List of blocks that hold the BAT (sector 0 from 0x004C to 0x01FF) */
	int *xbat_list;					/* List of extended BAT blocks starting with 109 */
	int *sbat_list;					/* List of SBAT blocks */
	int sbat_list_max;				/* Debug_ Number of elements in sbat_list */
	int bat_size, sbat_size, xbat_size;		/* Number of elements in BAT, SBAT, XBAT */
	int xbat_start, sbat_start;		/* Start blocks of XBAT and SBAT */
	int root_block;					/* Block index that holds the root property */
	poi_prop_type *proplist;		/* All properties from the root property chain */
	int props;						/* Number of properties in proplist */
	int sstream_start;				/* Start block of the SBAT-indexed mini stream */
	int sstream_size;				/* Total size of the SBAT-indexed mini-stream */
	unsigned char is_open;			/* Set to 1 if successfully openend */
}
poifs_type;

poifs_type poifs;					/* Allow one POI file system */

gchar poi_report[32768];

///#define ADDPOIRPT(C)	{ if (strlen(poi_report)<32000) strcat (poi_report, C); }

void ADDPOIRPT (gchar *mumble)
{
int i;


	if (strlen(poi_report) + strlen(mumble) < 32000)
	{
		for (i=0; i<strlen(mumble); i++)			/* Might as well try a clean-up */
			if (mumble[i] > 127)
				mumble[i] = '?';

		strcat (poi_report, mumble);
	}
}



/* Read a block (512 bytes, typically, 64 bytes for short blocks) in a random-access fashion */

int readblock (FILE *FP, unsigned char *block, int blockidx, int blksize)
{
long filepos;
int n;

	filepos = blksize * (blockidx+1);
	fseek (FP, filepos, SEEK_SET);
	n=fread (block, blksize, sizeof (unsigned char), FP);
	if (n==blksize)
		return 0;
	else
		return -1;
}
	


/* Read the property data if the n-th property (0<=n<=3) of a raw 512-byte block.
	NOTE -- the block can be of arbitrary length, as long as the index stays within bounds */

poi_prop_type get_property (unsigned char *block, int propidx)
{
int idx,i,j;
poi_prop_type prop;

	memset (&prop, 0, sizeof(prop));

	idx = 128*propidx;					/* Property blocks are 128 bytes long */
	j=0;
	for (i=idx; i<idx+64; i+=2)			/* Unicode chars are two-byte. We extract only the first (i.e., ASCII) */
		prop.name[j++] = block[i];
	prop.name[j]=0;

	prop.prop_type = block[idx+0x42];
	memcpy (&prop.prev_prop,   &(block[idx+0x44]), sizeof(int));
	memcpy (&prop.next_prop,   &(block[idx+0x48]), sizeof(int));
	memcpy (&prop.child_prop,  &(block[idx+0x4C]), sizeof(int));
	memcpy (&prop.state_flags, &(block[idx+0x60]), sizeof(int));
	memcpy (&prop.start_blk,   &(block[idx+0x74]), sizeof(int));
	memcpy (&prop.size,        &(block[idx+0x78]), sizeof(int));
	return prop;
}



void list_prop (poi_prop_type prop)
{
gchar buf[64];

	strcpy (buf, (gchar *)prop.name);				/* Format the string to length 32, padded w/ blanks */
	strncat (buf, "                                  ", 33); buf[33]=0;
	if (prop.prop_type==5)
		dp (2, "R");
	else if (prop.prop_type==2)
		dp (2, "F");
	if (prop.prop_type==1)
		dp (2, "D");
	dp (2, " %6d  %34s", prop.size, buf);
	dp (2, "   (Prev=%d, Next=%d, Child=%d, Start=%d)\n",
		prop.prev_prop, prop.next_prop, prop.child_prop, prop.start_blk);

}


void list_sbat_index ()
{
int i;

	dp (2, "POIFS: This is the SBAT (total %d):", poifs.sbat_list_max);
	for (i=0; i<poifs.sbat_list_max; i++)
	{
		if ((i & 7) ==0) dp (2, "\n[%4d]  ", i);
		dp (2, "    %6d", poifs.sbat_list[i]);
	}
	if ((i & 7) ==0) dp (2, "\n");
}


/* The BAT covers blocks 000 to 108, everything above 108 is in the XBAT. 
	The XBAT is a simple linked list (the last INT of each XBAT sector is the index
	of the next XBAT sector), so each XBAT sector contributes 127 blocks that hold
	128 BAT entries each. We just need to read the chain of these */


int read_xbat_chain (FILE *FP, int *xbatptr, int start_blk)
{
int nextblk, blockcnt;
int bat[128];			/* One block of block allocation tables */
int xbat_entries;


	blockcnt = 1;												/* Reflect the first sector read below */
	xbat_entries=0;
	readblock (FP, (unsigned char *)bat, start_blk, 512);		/* First XBAT block */
	dp (3, "POI XBAT chain: Reading XBAT block %d\n", start_blk);
	memcpy (xbatptr, bat, 127*sizeof(int));						/* Copy the data */
	xbatptr += 127;												/* and advance the pointer */
	xbat_entries += 127;

	nextblk = bat[127];					/* Point to next XBAT block in chain, or -2 for end of chain */

	while (nextblk > 0)
	{
		blockcnt++;
		readblock (FP, (unsigned char *)bat, nextblk, 512);		/* Next XBAT block */
		dp (3, "POI XBAT chain: Reading XBAT block %d\n", nextblk);
		memcpy (xbatptr, bat, 127*sizeof(int));						/* Copy the data */
		xbatptr += 127;												/* and advance the pointer */
		xbat_entries += 127;
		nextblk = bat[127];					/* Point to next XBAT block in chain, or -2 for end of chain */
		if (nextblk == -2)
			dp (3, "POI XBAT chain: End-Of-Chain marker reached. \n");
	}

	dp (2, "POI XBAT chain: %d XBAT blocks read\n", blockcnt);

	return blockcnt;
}


/* This function reads a file of conescutive sectors (a chain) as given by the 
	entries in the BAT array. A start block is given, and each BAT entry that
	corresponds to the sector contains the next sector in the chain. */


int read_chain (FILE *FP, unsigned char **bufptr, int size, int start_blk, int dumpsbat)
{
int nextblk, bat_idx, cached_bat, blockcnt;
int bat[128];			/* One block of block allocation tables */
int datalen;
unsigned char *buf;
gchar mumble[512];
int ministream_blks[2048];
int ministream_idx,i;


	/* Allocate the memory */

	if (size==0)					/* Initiate a dummy read to determine size */
		buf = NULL;
	else
		buf = calloc (size + 1024, sizeof(unsigned char));		/* Ensure we have more space than needed */

	datalen=0; blockcnt=0;
	*bufptr = buf;					/* Pass the pointer value back to caller */

	nextblk = start_blk;
	cached_bat = -1;
	ministream_idx=0;

	/* The first block is given by start_blk (now also nextblk), does not need a look-up,
		so read it immediately */

	if (buf)								/* Maintain dummy read if buf==NULL */
	{
		readblock (FP, buf, nextblk, 512);	/* Read next chunk of data */
		buf += 512;							/* Advance the data target pointer */
	}
	datalen += 512;	blockcnt++;				/* Keep track of how much we have read */
	dp (3, "POI chain: Read initial DATA block in chain %d\n", nextblk);
	if (dumpsbat && (ministream_idx<2000)) ministream_blks[ministream_idx++] = nextblk;

	/* To enter the chain, we need to look-ahead the next chained block.
		If this turns out to be -2, skip all of the rest. Note that the start block
		is arbitrary and can be part of the XBAT chain. 
		The BAT block is guaranteed not to be cached, so we initially need to read that one, too. */

	bat_idx = nextblk/128;
	if (bat_idx >= 109)						/* This is a case for the XBAT */
	{
		bat_idx = poifs.xbat_list[bat_idx-109];
	}
	else
		bat_idx = poifs.bat_list[bat_idx];	/* Index into bat_list for the currently read block */

	readblock (FP, (unsigned char *)bat, bat_idx, 512);
	cached_bat = bat_idx;				/* New BAT block cached */
	nextblk = bat[nextblk % 128];			/* Look up next data block in chain */

	/* Read the chain until an end-of-chain marker (-2) is found. */

	while (nextblk >= 0)			/* A negative value indicates the end of chain */
	{
		bat_idx = nextblk/128;
		if (bat_idx >= 109)						/* This is a case for the XBAT */
		{
			dp (3, "POI chain: Looking up index %d of the XBAT", bat_idx);
			bat_idx = poifs.xbat_list[bat_idx-109];
			dp (3, ", which is in block %d\n", bat_idx);
		}
		else
			bat_idx = poifs.bat_list[bat_idx];	/* Index into bat_list for the currently read block */

		if (bat_idx != cached_bat)				/* Past current BAT block: We did not (yet) read that block */
		{
			dp (3, "POI chain: Reading BAT block %d to look up data block %d\n", bat_idx, nextblk);
			readblock (FP, (unsigned char *)bat, bat_idx, 512);
			cached_bat = bat_idx;				/* New BAT block cached */
		}

		if (buf)								/* Maintain dummy read if buf==NULL */
		{
			readblock (FP, buf, nextblk, 512);	/* Read next chunk of data */
			buf += 512;							/* Advance the data target pointer */
		}
		datalen += 512;	blockcnt++;				/* Keep track of how much we have read */
		dp (3, "POI chain: Read DATA block %d\n", nextblk);
		if (dumpsbat && (ministream_idx<2000)) ministream_blks[ministream_idx++] = nextblk;

		nextblk = bat[nextblk % 128];			/* Look up next data block in chain */
	}

	if (nextblk != -2)
	{
		dp (1, "POI chain: WARNING: END OF BLOCK IS %d\n", nextblk);
		sprintf (mumble, "POI warning: chain with start %d rain into end marker of %d\n", start_blk, nextblk);
		ADDPOIRPT (mumble);
	}
	dp (2, "POI chain: Total of %d data blocks (%d bytes) read, starting at block %d\n", blockcnt, datalen, start_blk);

	if (dumpsbat) 
	{
		dp (1, "\nMINISTREAM: %d BLOCKS ---------------------------------\n", ministream_idx);
		for (i=0; i<ministream_idx; i++)
		{
			if ((i & 7) == 0)
				dp (1, "\n[%4d]:    ", i);
			dp (1, "  %6d", ministream_blks[i]);
		}
		if (ministream_idx==2000)
			dp (1, "\nLONG MINISTREAM, READ PROBABLY ABORTED.-------------------\n");
		else
			dp (1, "\nEND MINISTREAM -------------------------------------------\n");
		//sleep (2);
	}


	return datalen;
}




/* This function reads a file of conescutive sectors (a chain) from the SBAT minichain.
	Here, each sector is only 64 bytes, and the SBAT blocks are indexed by poifs.sbat_list[]
	Functionally, this is identical to the read_chain function above. 
	The helper function "load_minichain_block" is needed to traverse the BAT to access
	the SBAT block which contains the block number. Since each SBAT block is 64 bytes,
	each full-size block holds 8 SBAT blocks. */

/* DOUBLE INDIRECTION IS MISSING. WE READ THE BAT INDEX, NOT THE BAT BLOCK !!! */

int load_minichain_block (FILE *FP, unsigned char *buf, int block)
{
int targblock, thisblk, nextblk, blockcnt;
int idx, bat_idx, cached_bat,i;
int batbuf[512];						/* To buffer one BAT block */
unsigned char data[512];				/* To buffer one data block */
gchar mumble[512];


	targblock = block/8;				/* Get the necessary 512-byte block number */
	idx = block % 8;					/* Index into 512-byte BAT block */
	blockcnt=0;
	thisblk = poifs.sstream_start;		/* This is our entry point into the BAT (usually block 3) */

	/* Which BAT block contains the starting block? load it. */

	bat_idx = thisblk / 128;
	if (bat_idx >= 109)						/* This is a case for the XBAT */
	{
		bat_idx = poifs.xbat_list[bat_idx-109];
	}
	else
		bat_idx = poifs.bat_list[bat_idx];	/* Index into bat_list for the currently read block */

	readblock (FP, (unsigned char *)batbuf, bat_idx, 512);
	cached_bat = bat_idx;					/* New BAT block cached */
	nextblk = batbuf[thisblk % 128];		/* Look up NEXT data block in chain */

	/* we need to work our way thorugh the BAT chain until we have reached
		the target block, whose number in the ministream chain is targblock */
	
	while ((blockcnt < targblock) && (thisblk>=0))		/* Advance until we found the BAT block */
	{
		bat_idx = nextblk/128;							/* Which BAT block contains this block? */
		if (bat_idx >= 109)
			bat_idx = poifs.xbat_list[bat_idx-109];
		else
			bat_idx = poifs.bat_list[bat_idx];

		if (bat_idx != cached_bat)				/* Past current BAT block: We did not (yet) read that block */
		{
			readblock (FP, (unsigned char *)batbuf, bat_idx, 512);
			cached_bat = bat_idx;				/* New BAT block cached */
		}
		thisblk = nextblk;
		nextblk = batbuf[thisblk % 128];	/* Look up NEXT data block in chain */
		blockcnt++;	
	}

	/* Once we found the BAT block which contains the SBAT block, load it.
		Then, finally, we have access to the ministream DATA. */

	if (thisblk < 0)
	{
		sprintf (mumble, "POI: SBAT ERROR. Ran into END-OF-BLOCK before reaching target block\n");
		ADDPOIRPT (mumble);
		dp (0, "%s", mumble);
	}
	else
	{
		readblock (FP, data, thisblk, 512);
		dp (2, "POI SBAT: Loaded BAT block %d which contains SBAT block %d\n", thisblk, block);
		memcpy (buf, &(data[idx*64]), 64);	/* And copy the 64 bytes that are relevant to buf */
		if (debuglevel >2)
		{
			dp (2, "     -------------------------------------------");
			for (i=0; i<64; i++)
			{
				if ((i & 7) ==0)
					dp (2, "\n     [%3d]  ",i);
				dp (2, "  %02X", data[idx*64+i]);
			}
			dp (2, "\n     -------------------------------------------\n");
		}

	}
	return blockcnt;

}


int read_minichain (FILE *FP, unsigned char **bufptr, int size, int start_blk)
{
int sblock, blockcnt, datalen;
unsigned char *buf;
gchar mumble[512];


	/* Allocate the memory */

	if (size==0)					/* Initiate a dummy read to determine size */
		buf = NULL;
	else
		buf = calloc (size + 1024, sizeof(unsigned char));		/* Ensure we have more space than needed */

	datalen=0; blockcnt=0;
	*bufptr = buf;					/* Pass the pointer value back to caller */

	/* Read the first SBAT block and the first data block; get the next block from BAT */

	sblock = start_blk;

	do
	{
		if (buf) 										/* Dummy read if buf==NULL */
		{
			load_minichain_block (FP, buf, sblock);		/* Load 64 bytes of mini-stream data */
			buf += 64;									/* Advance the data pointer */
		}
		datalen += 64; blockcnt++;
		sblock = poifs.sbat_list[sblock];			/* Look up next block in chain */
	}
	while (sblock >= 0);							/* Stop when end-of-chain marker */

	if (sblock != -2)
	{
		dp (2, "POI minichain: WARNING: END OF BLOCK MARKER IS %d\n", sblock);
		sprintf (mumble, "POI warning: Minichain block with start %d ran into end marker of %d\n", start_blk, sblock);
		ADDPOIRPT (mumble);
	}
	dp (2, "POI minichain: Total of %d data blocks (%d bytes) read\n", blockcnt, datalen);

	return datalen;
}



/* Higher-level functions to read files (data streams). We need:
	--  a start block. This can either be a SBAT block or a BAT block
	--  a size. If the size is 4095 or less, it's a SBAT stream. Otherwise, a BAT stream.
	We return the raw data in buf, and the function returns the number of bytes read.
	Note that the caller is responsible for disposing of the buffer 

	Moreover, we conveniently create a series of helper functions that read
	typed variables (int, float, double etc) from the mini stream.
*/

int poi_read_file (FILE *FP, unsigned char **bufptr, int size, int start_blk)
{
int numbytes;

	if (size <= 4095)
		numbytes = read_minichain (FP, bufptr, size, start_blk);
	else
		numbytes = read_chain (FP, bufptr, size, start_blk,0);

	return numbytes;
}


unsigned short poi_read_short (FILE *FP, int start_blk)
{
unsigned char *buf;
unsigned short result;

	poi_read_file (FP, &buf, sizeof (short), start_blk);
	memcpy (&result, buf, sizeof (short));
	free (buf);
	return result;
}


int poi_read_int (FILE *FP, int start_blk)
{
unsigned char *buf;
int result;

	poi_read_file (FP, &buf, sizeof (int), start_blk);
	memcpy (&result, buf, sizeof (int));
	free (buf);
	return result;
}


float poi_read_float (FILE *FP, int start_blk)
{
unsigned char *buf;
float flt;

	poi_read_file (FP, &buf, sizeof (float), start_blk);
	memcpy (&flt, buf, sizeof (float));
	free (buf);
	return flt;
}

float poi_read_double (FILE *FP, int start_blk)
{
unsigned char *buf;
double flt;

	poi_read_file (FP, &buf, sizeof (double), start_blk);
	memcpy (&flt, buf, sizeof (double));
	free (buf);
	return flt;
}


/* cxd_open is quite a crucial function. It opens the POI filesystem and reads
	the header info. On top of that, it reads:
	* the BAT index and the XBAT index
	* the SBAT index
	* the property list (all of them) 

	TODO: We read the blocksize, but never use it (constant 512 or 64), should be fixed!
	
*/


FILE *cxd_open (const char* filename)
{

FILE* FP;
gchar mumble[512];
unsigned char header[512];
unsigned char *buf;
int numelems;
int cxdversion;
int i;
unsigned short u;


	poifs.is_open = 0;
	FP = fopen (filename, "rb");
	if (!FP) return NULL;

	numelems = fread (header, sizeof (char), sizeof(header), FP);
	if (numelems != 512)
	{
		dp (0, "Error: Read %d header elements, should be 512\n", numelems);
		fclose (FP);
		return NULL;
	}

	/* Check the magic ID */

	if ( (header[0] != 0xD0) || (header[1] != 0xCF) || (header[2] != 0x11) || (header[3] != 0xE0) )
	{
		statusbar_display ("Incorrect signature for cxd file. Aborted.");
		fclose (FP);
		return NULL;
	}

	snprintf (mumble, 500, "Opened and validated POI file %s\n\n", filename);
	ADDPOIRPT (mumble);

	/* Can get some more info here... */

	memcpy (&cxdversion, &(header[0x18]), sizeof (int));
	snprintf (mumble, 500, "POI file version (major:minor) %02X:%02X\n", cxdversion >> 16, cxdversion & 0xFFFF);
	ADDPOIRPT (mumble);
	dp (2, "POI: POI file version (major:minor) %02X:%02X\n", cxdversion >> 16, cxdversion & 0xFFFF);
	if ((cxdversion >> 16) != 3)
	{
		snprintf (mumble, 500, "Warning: Trying to read version %d cxd file as I would version 3.\nTHIS IS PROBABLY NOT GOOD.\n", cxdversion >> 16);
		ADDPOIRPT (mumble);
	}

	u = header[0x1E] + (header[0x1F] << 8);
	poifs.blksize = 1 << u;
	u = header[0x20] + (header[0x21] << 8);
	poifs.sblksize = 1 << u;
	dp (2, "POI: Block size of %d bytes (mini blocks %d bytes)\n", poifs.blksize, poifs.sblksize);

	memcpy (&poifs.bat_size, &(header[0x2C]), sizeof (int));
	dp (2, "POI: %d elements in BAT array\n", poifs.bat_size);
	memcpy (&poifs.sbat_start, &(header[0x3C]), sizeof (int));
	memcpy (&poifs.sbat_size, &(header[0x40]), sizeof (int));
	dp (2, "POI: %d elements in SBAT array, starts at %d\n", poifs.sbat_size, poifs.sbat_start);
	memcpy (&poifs.xbat_start, &(header[0x44]), sizeof (int));
	memcpy (&poifs.xbat_size, &(header[0x48]), sizeof (int));
	dp (2, "POI: %d elements in XBAT array, starts at %d\n", poifs.xbat_size, poifs.xbat_start);

	memcpy (&poifs.root_block, &(header[0x30]), sizeof (int));
	dp (2, "POI: Block index of property table: 0x%08X \n", poifs.root_block);

	/* Read the BAT block list. There are 109 entries in the BAT plus several in the XBAT.
	   We know the number of xbat entries, and this allows us to allocate the memory  */

	memcpy (poifs.bat_list, &(header[0x4C]), 109*sizeof(int));			/* Get the BAT index integers */
	poifs.xbat_list = calloc (poifs.xbat_size*127, sizeof (int));		/* Dynamically allocate the XBAT */
	numelems = read_xbat_chain (FP, poifs.xbat_list, poifs.xbat_start);
	dp (2, "%d XBAT entries read\n", numelems);

	/* Read the SBAT index chain. The SBAT is most always used, because it is used to
		store files of size 4095 and less (such as commens, real & integer parameters etc).
		The SBAT is a BAT chain like every other and yields a list of INTs that are the index
		to the SBAT blocks. Note that we can read the SBAT here, but for accessing the
		actual ministream, we need to get the stream info from the root property (below) */

	numelems = read_chain (FP, &buf, 0, poifs.sbat_start,1);			/* Dummy read to assess length of SBAT index */
	dp (2, "POI SBAT: Read %d entries\n", numelems/sizeof(int));

	poifs.sbat_list = calloc (numelems/sizeof(int), sizeof(int));
	numelems = read_chain (FP, (unsigned char **)(&poifs.sbat_list), numelems, poifs.sbat_start,0);
	dp (2, "POI SBAT: Index has %d blocks for a total SBAT storage size of %d\n", numelems/sizeof(int), 64*numelems/sizeof(int));
	poifs.sbat_list_max = numelems/sizeof(int);
	list_sbat_index();
	//sleep (2);

	/* Read the property list */

	numelems = read_chain (FP, &buf, 0, poifs.root_block,0);			/* Dummy read to assess length of property table */
	dp (2, "POI Properties: %d properties read\n", numelems/0x80);
	numelems = read_chain (FP, &buf, numelems, poifs.root_block,0);	/* Now do the real read with known size */
	numelems /= 0x80;												/* Each property is 128 bytes long */
	poifs.props = numelems;											/* Make number of properties available */
	poifs.proplist = calloc (numelems, sizeof (poi_prop_type));		/* Allocate memory for the clear-text properties */
	for (i=0; i<numelems; i++)
	{
		poifs.proplist[i] = get_property (buf, i);
		if (i==0)										/* The root property contains mini stream info */
		{
			poifs.sstream_start = poifs.proplist[i].start_blk;
			poifs.sstream_size = poifs.proplist[i].size;
		}

		dp (2, "%4d:   ", i);
		list_prop (poifs.proplist[i]);
	}

	snprintf (mumble, 500, "POIFS contains %d properties.\n", numelems);
	ADDPOIRPT (mumble);
	snprintf (mumble, 500, "POIFS Mini-stream info:\nMini-stream of size %d starts at block %d,\nwith SBAT starting at %d and having %d blocks\n", poifs.sstream_size, poifs.sstream_start, poifs.sbat_start, poifs.sbat_size);
	ADDPOIRPT (mumble);
	numelems = read_chain (FP, &buf, 0, poifs.sstream_start,1);			/* Dummy read to dump SBAT bocks */

	poifs.is_open = 1;			/* YAY!! */

	/* Note: What we are really looking for is the "Bitmap" property. For example, a Bitmap
		property occurs at block 9 (0x1400) with the block (at 0x1474) given as 0x12. This resolves to
		byte offset 0x2600, which is where a bitmap really starts. */


	return FP;
}



int poi_parse_props (FILE *FP, image_type *img)
{
int numfields;						/* We count all properties that start with "Field" */
int field_count;					/* This is the content of the file property "Field Count" */
unsigned short has_image;			/* This is the content of the file property "File Has Image" */
int i;
gchar mumble[512];
gchar propname[64];					/* merely to get rid of those pesky signedness warnings */
unsigned char *buf;
poi_prop_type pr;					/* Need a convenient shorthand for poifs.proplist[i] */
int imwidth, imheight, imdepth, imcolors;
double ddummy;						/* Why the %&@!* do they store image sizes as DOUBLE??? */
TYPE_SELECTOR imgtype;
int bpp;							/* Bytes per pixel, later used to determine type */


	imwidth=0; imheight=0; imdepth=0; imcolors=1;
	numfields=0;
	has_image = 0;

	if (poifs.is_open)
	{
		for (i=0; i<poifs.props; i++)			/* Scan all properties */
		{
			pr = poifs.proplist[i];
			strcpy (propname, (gchar *)pr.name);
			if (!strncasecmp (propname, "Field Count",12))		/* This has to be first to not interfere with "Field nnn" */
			{
				field_count = poi_read_int (FP, pr.start_blk);
			}
			else if (!strncasecmp (propname, "File Has Image",14))
			{
				has_image = poi_read_int (FP, pr.start_blk);
			}
			else if (!strncasecmp (propname, "Field",5))
			{
				if ((propname[6] >='0') && (propname[6] <='9'))		/* Has the format Field nnn */
				{
					numfields++;
				}
			}
			else if (!strncasecmp (propname, "Comments",8))
			{
				sprintf (mumble, "\nWe are in luck! File has comments. Here they are:\n--------------------------------------------------\n");
				ADDPOIRPT (mumble);
				poi_read_file (FP, &buf, pr.size, pr.start_blk);
				ADDPOIRPT ((gchar*)buf);
				strcpy (mumble, "\n--------------------------------------------------\n\n");
				ADDPOIRPT (mumble);
				if (debuglevel>1) fprintf (stderr, buf);
			}
			else if (strstr (propname, "_Image_Width"))
			{
				ddummy = poi_read_double (FP, pr.start_blk);
				imwidth = MAX (imwidth, (int)(ddummy));
			}
			else if (strstr (propname, "_Image_Height"))
			{
				ddummy = poi_read_double (FP, pr.start_blk);
				imheight = MAX (imheight, (int)(ddummy));
			}
			else if (strstr (propname, "_Image_Depth"))
			{
				ddummy = poi_read_double (FP, pr.start_blk);
				imdepth = MAX (imdepth, (int)(ddummy));
			}
			else if (strstr (propname, "_Color_Components"))
			{
				ddummy = poi_read_double (FP, pr.start_blk);
				imcolors = MAX (imcolors, (int)(ddummy));
			}
		}
	}
	else
		return -1;

	bpp=0;
	if (imdepth==0) imdepth=8;		/* Not found or not set -> make default depth of 1 byte */
	while ((imdepth % 8) !=0) imdepth++;		/* Round up to next byte */
	bpp = imdepth/8;
	sprintf (mumble, "%d properties scanned. %d fields are reported. File has image? ", poifs.props, field_count);
	ADDPOIRPT (mumble);
	if (has_image) strcpy (mumble, "yes."); else strcpy (mumble, "no [what??]");
	ADDPOIRPT (mumble);
	sprintf (mumble, "\nTotal of %d ´FIELD´ entries\n", numfields);
	ADDPOIRPT (mumble);
	sprintf (mumble, "We strongly suspect that this is an image of %dx%dx%d voxels with %d bits/voxel and %d color channels\n", imwidth, imheight, numfields, imdepth, imcolors);
	ADDPOIRPT (mumble);

	imgtype=B_CHAR;
	if (bpp==2)
		imgtype=B_SHORT;
	else if (bpp==3)
		imgtype=B_RGB;
	else if (bpp==4)
		imgtype=B_FLOAT;

	numfields = field_count;			/* Not sure, but maybe we can trust this one more than the plain-text count? */

	/* KLUDGE! We still don't know how bpp and color components relate to the actual bytes per slice */

	if (allocate_image (img, imgtype, imwidth, imheight, 2+3*numfields*imcolors)==-1)
		fatal_memerr();

	return 0;
}


/* Whew. Made it to this point. Here, we actually import image data.
	The following properties form a tree:
			Field nnn  --> i_Image1  --> Bitmap 1
							--> Details --> width, height, etc.
	At first glance, the tree structure seems rhater inconsistent, so we try to kludge
	the image together by reading each Field property and chaining down to its bitmap.
	Then we stuff the bitmap into one slice.

*/

void poi_import_image (FILE *FP, image_type *img)
{
int i,j,e,slicesize;
gchar mumble[512];
gchar propname[64];					/* merely to get rid of those pesky signedness warnings */
poi_prop_type pr;					/* Need a convenient shorthand for poifs.proplist[i] */
poi_prop_type im_pr, bmp_pr;		/* Image and Bitmap child properties */
int fieldnum;
unsigned char *imgbuf;				/* Generic image type, will be corrected unpon image insertion */
unsigned char *destptr;



	if (!poifs.is_open)
		return;

	e = DATALEN(img->imgtype);
	slicesize = e*img->xmax*img->ymax;
	destptr = img->data;

	for (i=0; i<poifs.props; i++)			/* Scan all properties */
	{
		pr = poifs.proplist[i];
		strcpy (propname, (gchar *)pr.name);
		if (!strncasecmp (propname, "Field",5))
		{
			if ((propname[6] >='0') && (propname[6] <='9'))		/* Has the format Field nnn */
			{
				sscanf (propname, "%s %d", mumble, &fieldnum);
				j = pr.child_prop;
				dp (2, "Found Field #%d with child property #%d\n", fieldnum, j);
				
				im_pr = poifs.proplist[j];
				strcpy (propname, (gchar *)im_pr.name);
				if (!strncasecmp (propname, "i_Image",7))
				{
					j = im_pr.child_prop;
					bmp_pr = poifs.proplist[j];
					strcpy (propname, (gchar *)bmp_pr.name);
					if (!strncasecmp (propname, "Bitmap",6))
					{
						/* Got it! Read the buffer and add it to the image */

						poi_read_file (FP, &imgbuf, bmp_pr.size, bmp_pr.start_blk);
						destptr = (unsigned char *)(img->data) + 2*slicesize*fieldnum;	/* still a KLUDGE! */
						memcpy (destptr, imgbuf, bmp_pr.size);
						free (imgbuf);
						sprintf (mumble, "Inserted Field %d as slice %d (copied %d bytes).\n", fieldnum, fieldnum, bmp_pr.size);
						ADDPOIRPT (mumble);
					}
					else
					{
						sprintf (mumble, "Field %d does not have a child named 'Bitmap'. Moving on.\n", fieldnum);
						ADDPOIRPT (mumble);
					}
					
				}
				else
				{

					/* Since the tree is sloppily implemented by Compix, let's try to brute-force the directory
						and try if we find the bitmap two down, as it is typically stored. */

					bmp_pr = poifs.proplist[i+2];
					strcpy (propname, (gchar *)bmp_pr.name);
					if (!strncasecmp (propname, "Bitmap",6))
					{
						/* Got it! Read the buffer and add it to the image */

						poi_read_file (FP, &imgbuf, bmp_pr.size, bmp_pr.start_blk);
						destptr = (unsigned char *)(img->data) + 2*slicesize*fieldnum;	/* still a KLUDGE! */
						memcpy (destptr, imgbuf, bmp_pr.size);
						free (imgbuf);
						sprintf (mumble, "Inserted Field %d as slice %d (copied %d bytes).\n", fieldnum, fieldnum, bmp_pr.size);
						ADDPOIRPT (mumble);
					}
					else
					{
						sprintf (mumble, "Field %d does not have a child named 'i_Image'. Moving on.\n", fieldnum);
						ADDPOIRPT (mumble);
					}
				}

			}
		}
	}

	make_minmax (img);
	img->uid = next_uid++;

	sprintf (mumble, "\nIMPORT COMPLETE. ENJOY.\n");
	ADDPOIRPT (mumble);

}



int cxd_read  (image_type* img, const char* filename)
{
FILE *FP;


	poi_report[0] = 0;

	FP = cxd_open (filename);
	if (FP == NULL)
		return -1;

	if (poi_parse_props (FP, &mainimg)) 
	{
		ShowLongMessage ((gchar *)poi_report, "CXD import failure!");
		return -1;
	}

	poi_import_image (FP, &mainimg);
	ShowLongMessage ((gchar *)poi_report, "CXD import log");
	fclose (FP);
	poifs.is_open=0;
	return 0;
}





/************************************************************************************

	Another worthy file format is the Nikon nd2 format.
	Little documentation exists, so it boils down to hacking, unless one wants
	to use the proprietary Nikon Windows-only <groan> dll.

	Apparently, a decent import filter built with Python exists:
	https://github.com/jimrybarski/nd2reader/tree/master/nd2reader
	which can be used to glean some information.

	Worth noting: The microscope output of one sample file has the raw and
	uncompressed image data starting at 0x0F000. Pixels are 16-bit big-endian,
	and multiple channels appear interlaced. In the run-up to 0x0F000, there is
	a lot of metadata, and we'll likely find the image size and so on.

	The nd2 can be identified by its header. We should read the first 16 bytes.
	The file begins with the sequence DA:CE:BE:0A:20:00:00:00 (long int), followed
	by (according to Jim Rybarski) followed by the data length (4 bytes) and
	the relative offset (4 bytes).

*************************************************************************************/




/************************************************************************************

	High-level module control

*************************************************************************************/




void exotic_file_import (gchar* filename)
{


	/* Test the header magic to find out fhat file type it is */
	/* TODO: test the HEADER MAGIC to find out... */

	if (strstr (filename, ".wim"))
	{
		if (wim_read (&mainimg, filename) < 0)
		{
			ShowMessage ("Read failed");
			return;
		}
	}
	else if (strstr (filename, ".cxd"))
	{
		if (cxd_read (&mainimg, filename) < 0)
		{
			ShowMessage ("Read failed");
			return;
		}
	}
	else
	{
		ShowMessage ("Unknown format");
		return;
	}

	/* Create an imlib image for display */

	make_minmax (&mainimg);
	cur_slice = 0;
	reset_zoomlevel ();
	make_imlib_buffer_from_main_image (cur_slice);
	prepare_display_window (filename);		/* Causes a configure event followed by an expose event */

	file_ok = 1;
	statusbar_display ("File loaded OK");
	strcpy (mainimg_name, filename);

	redraw (0);			/* Just in case, schedule an expose event */
	
}





void module_import_exotic ()
{
gchar* selectedfile;

	ChooseFile("Select file to import:", window1, 0, &selectedfile);
	if (selectedfile)
	{
		dp (1, "File selected: %s\n", selectedfile);
		exotic_file_import (selectedfile);
		g_free (selectedfile);
	}

}










/**************************************************************

	This is the infrastructure we use to register with the 
	main program. The pointer "menufunction" points to the 
	main entry (the function to be called when the menu is 
	activated). Initmodule is the one function that the 
	main program calls (the name is fixed!) to initialize
	the module and to register the module with the menu.

***************************************************************/

void (*menufunction1)();









void initmodule ()
{


	dp (1, "Exotic file formats module: Initializer called\n");
	
	/* Register this module with the main menu, add under the "File" menu */

	menufunction1 = &module_import_exotic;
	dso_menu_register ("Import exotic format", menufunction1);

}


