
/*
 *  FFT: 2D Fourier Transform and related functions
 */
 

/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */

 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <fftw3.h>

#include "defs.h"
#include "miscop.h"
#include "application.h"
#include "imgbuf.h"
#include "functions.h"
#include "secondwnd.h"
#include "gtkhelpers.h"
#include "graphdsp.h"
#include "dlgbox.h"
#include "viz.h"
#include "fft.h"


/* Redefine fftw 2's fftw_real for fftw3: */

#ifndef fftw_real

#define fftw_real	double

#endif

/* In addition, the struct containing re and im no longer exists. Use
   these macros instead: */

#define c_re(C)	((C)[0])
#define c_im(C)	((C)[1])







/* Swap quadrants 1/3 and 2/4 to get the zero frequency into the center.
	Swapping quadrants is done in-place and can be used generally
	for all image types

	NOTE: Efficiency gain possible -- we can copy full strips of xm/2 length.
*/


void fft_swap_quadrants (image_type *img)
{
int x, y, z, xm, ym;
int i, j, e;
unsigned char buf[8];
unsigned char *p;
long sidx, didx;

	xm = img->xmax;
	ym = img->ymax;
	e = DATALEN(img->imgtype);
	p = img->data;

	for (z=0; z < img->zmax; z++)
	{
		for (y=0; y<ym; y++)
			for (x=0; x<xm; x++)
			{
				if (x<xm/2) i=x+xm/2; else i=x-xm/2;
				if (y<ym/2) j=y+ym/2; else j=y-ym/2;

				sidx = (x + xm*(y + ym*z))*e;
				didx = (i + xm*(j + ym*z))*e;
				memcpy (buf, &(p[sidx]), e);
				memcpy (&(p[sidx]), &(p[didx]), e);
				memcpy (&(p[didx]), buf, e);

			}
	}
}





/* Windowing function. Return a 2D window value for coordinates x and y.
	TODO: Can't we link this to the more general, function in functions2.c?? 
	Probably... we should really consider this.

	Note: The window functions are kept symmetrical (as opposed to separated).
	This means that the drop-off for a rectangular image is higher along the long
	edge. A symmetrical function is simply based on the normalized radius r, e.g.,

		a + b cos (r * pi)

	whereas a separated window uses two directional functions, e.g.,

		a + b cos (pi y/ym) cos (pi x/xm)

*/


double fft_windowfn (WINDOW_TYPE window, int x, int y,int xm, int ym)
{
double a,b,w,rmax,r;

	if (!window) return 1.0;

	rmax = SQR(xm/2) + SQR(ym/2);		/* Squared distance to corner */
	r = SQR(x - xm/2) + SQR(y-ym/2);	/* Squared distance of pixel from center */

	switch (window)
	{
		case W_WELCH:			/* Parabola, drops to almost-zero at the edges */
			w = 1.0 - 0.98*r/rmax;
			break;
		case W_HAMMING:			/* Hamming: Drops to 0.08 at the image edges */
			a = 0.54; b=0.46;
			w = a + b*cos(PI*sqrt(r/rmax));
			break;
		case W_HANN:			/* Hann: Drops to zero, but we prevent it from reaching pi; drops to 0.025 */
			a = 0.5; b=0.5;
			w = a + b*cos(0.9*PI*sqrt(r/rmax));
			break;
		case W_GAUSS:			/* Raised Gaussian with sigma=0.5 and sigma^2=0.25 (quite extreme) */
			a = (float)x-((float)xm+1.0)/2.0;
			b = (float)y-((float)ym+1.0)/2.0;
			w = 0.05+0.95*exp(-r / (0.25*rmax) );
			break;
		default:
			w=1.0;
	}

	return w;

}


/************************************************************

	FFT converts the image passed as img to its 2D
	Fourier transform. Only slice z is transformed.
	The transformed image is passed back through img,
	it has two slices and is of type B_FLOAT
   
*/


/* The following FFT-based functions use the fftw library (www.fftw.org).
   This means that you need fftw3 installed on your computer. 
   
   TODO: At this time, FFT only works on grayscale images. For RGB,
   we'd probably have to create a 6-slice FFT stack that needs to be
   appropriately treated in ifft as well.
   This would be easier if we had a generalized format with
   n panes of any type. Oh, well... TODO.  */



void fft2d (image_type* img, int z)
{
image_type fftimg;	/* Temporary storage of transformed image */
fftw_complex *out;	/* typed pointer to data area */
fftw_complex *in;	/* typed pointer to complex input */
float* p1;			/* Typed pointer to result image data area */
int xm,ym;			/* Power-of-two dimensions */
int i,j,x,y,imag;
float tempr, tempi;

fftw_plan plan;




	xm = img->xmax; ym=img->ymax;
	
	/* Get memory for transformed image. Two slices contain
	   real part and imaginary part */
	
	if (allocate_image (&fftimg, B_FLOAT, xm,ym,2)==-1)
		fatal_memerr();
	p1 = fftimg.data;

	/* Allocate complex arrays for FFT input and output. Although
		FFTW3 has a r2c conversion, there was some trouble and I 
		reverted to c2c in spite of its lower efficiency. */
		
	out = (fftw_complex*) fftw_malloc (2*xm*ym*sizeof (fftw_complex));
	if (out==0) fatal_memerr();
	in = (fftw_complex*) fftw_malloc (2*xm*ym*sizeof (fftw_complex));
	if (in==0) fatal_memerr();

	/* Since we used malloc instead of calloc, we zero all unused components
		and fill the real part of the input buffer with image data */

	for (y=0; y<ym; y++)
		for (x=0; x<xm; x++)
		{
			c_re(out[(y + ym*x)]) = 0.0;	/* Real component */
			c_im(out[(y + ym*x)]) = 0.0;	/* Imaginary component */
			c_im(in[(y + ym*x)]) = 0.0;		/* Imaginary component of input */
			c_re(in[(y + ym*x)]) = readbuf_flt (img,x,y,z);
		}


	dp (1, "FFT: New image size %d x %d\n",xm,ym);

	/* fftw requires us to create a "plan" of the planned transforms.
		In higher debug levels, also output the plan info. */
	
	plan = fftw_plan_dft_2d (xm,ym, in, out, FFTW_FORWARD, FFTW_ESTIMATE);
	if (debuglevel > 2)
		fftw_print_plan (plan);


	/* perform the actual transform - call to the library */
	
	fftw_execute (plan);


	/* Separate real and imag part; at the same time swap quadrants 
	   1/3 and 2/4 to get the zero frequency into the center.
	   Finally, we normalize the image at this point - I prefer
	   normalization after the forward transform as it gives more meaningful
	   FFT values (i.e. those that correspond with e.g. the image average) */
	
	imag = xm*ym;
	for (y=0; y<ym; y++)
		for (x=0; x<xm; x++)
		{
		
			if (x<xm/2) i=x+xm/2; else i=x-xm/2;
			if (y<ym/2) j=y+ym/2; else j=y-ym/2;
		
			tempr = c_re(out[y+ym*x]);			/* Real component */
			tempi = c_im(out[y+ym*x]);			/* Imaginary component */
			/* tempr = sqrt (SQR(tempr) + SQR(tempi)); */
			p1[i + xm*j] = tempr / (xm*ym);
			p1[i + xm*j + xm*ym] = tempi / (xm*ym);
			
		}

	
	/* Hand back via img */
	
	/* Done - the plan is no longer needed, nor are the intermediate arrays */

	fftw_destroy_plan (plan);
	fftw_free (out);
	fftw_free (in);
	handback (img, &fftimg);

}



void ifft2d (image_type* img)
{
image_type ifftimg;	/* Temporary storage of transformed image */
fftw_complex *in;	/* typed pointer to complex fft data area */
fftw_complex *out;	/* typed pointer to complex ifft data area */
float* p1;			/* Typed pointer to result image data area */
int xm,ym;			/* Power-of-two dimensions */
int i,j,x,y;

fftw_plan plan;


	if ((img->imgtype!=B_FLOAT) || (img->zmax!=2))
		ShowMessage ("Error: This image is not a frequency image");

	xm = img->xmax; ym = img->ymax;
	out = (fftw_complex*) fftw_malloc (2*xm*ym*sizeof (fftw_complex));
	if (out==0) fatal_memerr();
	in = (fftw_complex*) fftw_malloc (2*xm*ym*sizeof (fftw_complex));
	if (in==0) fatal_memerr();

	/* fftw requires us to create a "plan" of the planned transforms */
	
	plan = fftw_plan_dft_2d (xm,ym, in, out, FFTW_BACKWARD, FFTW_ESTIMATE);
	if (debuglevel > 2)
		fftw_print_plan (plan);

	/* Get memory for transformed image. Two slices contain
	   real part and imaginary part, ideally the imaginary part 
	   should be zero */
	
	if (allocate_image (&ifftimg, B_FLOAT, xm,ym,2)==-1)
		fatal_memerr();
	p1 = ifftimg.data;

	dp (1, "IFFT: New image size %d x %d\n",xm,ym);

	/* Fill the new buffer *P with data, pad with zeroes.
		Also, fill in sequence real - imag - real - imag ...
		Finally, swap quadrants in this same operation. The zero
		frequency point must be in the top left corner */
	
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			if (x<xm/2) i=x+xm/2; else i=x-xm/2;
			if (y<ym/2) j=y+ym/2; else j=y-ym/2;
			c_re(in[(j + ym*i)]) = readbuf_flt (img,x,y,0);	/* Real component */
			c_im(in[(j + ym*i)]) = readbuf_flt (img,x,y,1);	/* Imag component */
			c_re(out[(j + ym*i)]) = 0;
			c_im(out[(j + ym*i)]) = 0;
		}


	/* perform the actual transform - call to the FFTW library */
	
	fftw_execute (plan);

	/* Transfer real and imag part; do not normalize the image
	   because normalization occurred during the forward transform. */
	
	for (y=0; y<ym; y++)
		for (x=0; x<xm; x++)
		{
			p1[x + xm*y] = c_re(out[y+ym*x]);
			p1[x + xm*y + xm*ym] = c_im(out[y+ym*x]);
		}

	
	/* Hand back via img */
	
	/* Done - the plan is no longer needed */

	fftw_destroy_plan (plan);
	fftw_free (out);
	fftw_free (in);
	handback (img, &ifftimg);

}



/* Discrete Hartley transform */



void dht2d (image_type* img, int z, int direction)
{
image_type fftimg;	/* Temporary storage of transformed image */
fftw_real *out;		/* typed pointer to data area */
fftw_real *in;		/* typed pointer to real input */
float* p1;			/* Typed pointer to result image data area */
int xm,ym;			/* Power-of-two dimensions */
int i,j,x,y;
fftw_real imgscale;

fftw_plan plan;


	xm = img->xmax; ym=img->ymax;
	
	/* Get memory for transformed image. Two slices contain
	   real part and imaginary part */
	
	if (allocate_image (&fftimg, B_FLOAT, xm,ym,1)==-1)
		fatal_memerr();

	p1 = fftimg.data;

	out = (fftw_real*) calloc (2*xm*ym, sizeof (fftw_real));
	if (out==0) fatal_memerr();
	in = (fftw_real*) calloc (2*xm*ym, sizeof (fftw_real));
	if (in==0) fatal_memerr();

	dp (1, "DHT: New image size %d x %d\n",xm,ym);

	/* fftw requires us to create a "plan" of the planned transforms.
		In higher debug levels, also output the plan info. */
	
	plan = fftw_plan_r2r_2d (xm,ym, in, out, FFTW_DHT, FFTW_DHT, FFTW_ESTIMATE);
	if (debuglevel > 2)
		fftw_print_plan (plan);


	/* Fill the new input buffer *in with data, pad with zeroes.
	   For the inverse transform, swap quadrants beforehand */
	
	if (direction==0)
	{

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				in[(y + ym*x)] = readbuf_flt (img,x,y,z);
			}
	}
	else			/* Inverse DHT */
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (x<xm/2) i=x+xm/2; else i=x-xm/2;
				if (y<ym/2) j=y+ym/2; else j=y-ym/2;
				in[(j + ym*i)] = readbuf_flt (img,x,y,0);
			}
	}
	
	/* perform the actual transform - call to the library */
	
	fftw_execute (plan);


	/* Postprocess the transform. For the forward DHT, swap the quadrants
	   so that the DC component is in the center. In addition,
	   normalize by image scale factor (in accordance with Bracewell,
		J.Opt.Soc.Am 1983; 73: 1832-1835, the normalization takes place
		in the forward transform) */
	
	if (direction==0)
	{
		imgscale = 1.0/(xm*ym);
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (x<xm/2) i=x+xm/2; else i=x-xm/2;
				if (y<ym/2) j=y+ym/2; else j=y-ym/2;
				p1[(j + ym*i)] = out[y + ym*x]*imgscale;
			}
	}
	else		/* Inverse DHT. Why don't we need to swap quadrants here??? */
	{
		for (y=0; y<ym; y++)
			for (x=0; x<xm; x++)
			{
				p1[y + ym*x] = out[y + ym*x];
			}
	}


	/* Hand back via img */

	/* Done - the plan is no longer needed, nor are the intermediate arrays */

	fftw_destroy_plan (plan);
	free (out);
	free (in);
	handback (img, &fftimg);

}


/* Menu interface for forward and inverse FFT and DHT */

void fft_main (int direction)
{

	undo_nexus (&mainimg);
	
	if (direction==2)		/* forward DHT */
		dht2d (&mainimg,cur_slice,0);
	else if (direction==3)		/* inverse DHT */
		dht2d (&mainimg,cur_slice,1);
	else if (direction==0)		/* Forward FFT */
		fft2d (&mainimg,cur_slice);
	else if (direction==1)		/* Inverse FFT */
		ifft2d (&mainimg);
	
	make_imlib_buffer_from_main_image (0);
	redraw (0);

}



/************************************************************************/



void fft_store (image_type* img, fftw_real* in, int xm, int ym, int z, int rgb)
{
unsigned char* cp;
short* sp;
float* fp;
int ch,x,y,idx;


	idx = img->xmax * img->ymax * z;
	cp=img->data; sp=img->data; fp=img->data;

	switch (img->imgtype)
	{
		case B_CHAR:
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					ch = (int) (0.5+in[y+ym*x]);
					if (ch<0) ch=0; else if (ch>255) ch=255;
					cp[x + img->xmax*y + idx] = ch;
				}
			break;
		case B_SHORT:
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					ch = (int) (0.5+in[y+ym*x]);
					sp[x + img->xmax*y + idx] = ch;
				}
			break;
		case B_FLOAT:
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
					fp[x + img->xmax*y + idx] = in[y+ym*x];
			break;
		case B_RGB:
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					ch = (int) (0.5+in[y+ym*x]);
					if (ch<0) ch=0; else if (ch>255) ch=255;
					cp[3*(x + img->xmax*y + idx)+rgb] = ch;
				}
			break;
	}

}


/* The filter functions represent the frequency response. The frequency
   spatial coordinates can be computed from x,y; and the function returns
   the attenuation factor. The parameters a,b,c,d may vary from filter to
   filter. */

/*  The return value is a 2x2 matrix of Fourier filter coefficients. 
	Since a Fourier filter may be
 	complex, we need to be able to perform the following operation:
 	Assume the matrix is	|  a -jb |
 							| jc   d |
 	and we multiply the original complex number x + jy with that matrix, 
 	then our result is
 		Real part: x' = ax + by
 		Imag part: y' = cx + dy

	In the case of a filter that only influences magnitude and not phase,
	we would set b=c=0 and a=d=H where H is the transfer function magnitude
	Another example. The first derivative is H(f) = -jf , so we change
	the phase.For this filter, we'd use a=0, b=f, c=-f, d=0
*/



ffilt_mat (*fft_filter_func) (int x, int y, float a, float b, float c, float d);


/* The neutral filter, in == out */

ffilt_mat neutral_fft_filter (int x, int y, float a, float b, float c, float d)
{
ffilt_mat H;

	H.A[0]=1.0; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=1.0;

	return H;
}


/* A homomorphic high-pass to remove illumination inhomogeneities.
   a is the cutoff frequency,
   b is the steepness
   c is the overall gain

   REALLY?? Steepness b? I'd say steepness is a, cutoff is b.
   Uhm, yes, it would actually appear that way.
*/

ffilt_mat fft_homomorphic_hp (int x, int y, float a, float b, float c, float d)
{
float r;
ffilt_mat H;

	r = sqrt (SQR(x) + SQR(y));
	r = c / (1.0 + exp(-a*(r-b)));
	H.A[0]=r; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=r;
	return H;

}


/* A simple high-pass to remove illumination inhomogeneities.
	Oppenheim filter, named after the example in "High Dynamic Range Imaging"
	by Reinhard et al (p. 327)
   a is the DC attenuation,
   b is the steepness		 */

ffilt_mat fft_oppenheim_hp (int x, int y, float a, float b, float c, float d)
{
float r;
ffilt_mat H;

	r = sqrt (SQR(x) + SQR(y));
	a/=100; b/=100;
	r = (1.0-a) + a*b*r/(1+b*r);
	H.A[0]=r; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=r;
	return H;

}


/* Low-pass filter with Gaussian frequency response.
   a the width sigma,
   b is the cutoff frequency (where the lp filtering begins)
   c is the overall gain */


ffilt_mat fft_gaussian_lp (int x, int y, float a, float b, float c, float d)
{
float r;
ffilt_mat H;

	if (a<0.1) a=0.1;				/* Kludge: Since we divide by a below, don't allow it to be zero */
	r = sqrt (SQR(x) + SQR(y));		/* Spatial frequency of the coordinate pair (x,y), really (u,v) */
	if (r<b) 						/* Did we reach the cutoff ? */
		r = c;								/* If not, constant gain of c */
	else
		r = c/ (exp(0.5*SQR((r-b)/a)));		/* If yes, let the gain drop off from c with the Gaussian distance */
	
	H.A[0]=r; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=r;
	return H;

}


/* High-pass filter with inverse Gaussian frequency response.
   a the width sigma,
   b is the cutoff frequency (where the hp filtering begins)
   c is the overall gain
   d is the filter offset.

	The function reverts a Gaussian blur G(f). Dividing by G(f) is
	impractical because of the low values of G at large f, so we extend
	the reciprocal by the complex conjugate to obtain

	W(f) = G*(f) / | G(f) |^2

	Now, the noise component is considered by adding an offset d to the
	denominator,

	W(f) = G*(f) / ( | G(f) |^2 + d^2)

	where d is -- ideally -- the magnitude of the noise PSD over the image PSD.
	In practice, d can be adjusted to approximate the noise/image component.

	NOTE THAT THIS IMPLEMENTATION IS NOT A REAL WIENER FILTER, BUT MERELY
	AN APPROXIMATION. A real Wiener filter would need to estimate noise and image
	PSD at all frequencies and use those.  See discussion in Chapter 3.2.3 of
	Advanced Biomedical Image Analysis.
 */


ffilt_mat fft_wiener (int x, int y, float a, float b, float c, float d)
{
float r;
ffilt_mat H;

	/* Start off with the Gaussian as used above */

	if (a<0.1) a=0.1;
	r = sqrt (SQR(x) + SQR(y));
	if (r<b) 
		r = c;
	else
		r = c/ (exp(0.5*SQR((r-b)/a)));

	/* Now compute the Wiener function */

	r = r/(r*r+d*d/100.0);
	
	H.A[0]=r; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=r;
	return H;

}




/* Low-pass filter with n-th order Butterworth response
	H = c / (1 + a*(R/R0)^2n)
   a is an asymmetry factor,
   b is the cutoff frequency R0 (where the lp filtering begins)
   c is the overall gain 
   d is the order n. Note that the floating-point parameter d is truncated to integer n.  */

/* Both Butterworth filters ***MUST*** default to a=1 */

ffilt_mat fft_butterworth_lp (int x, int y, float a, float b, float c, float d)
{
float r;
int n;
ffilt_mat H;

	n = (int)d;
	if (a<0.1) a=0.1;
	r = sqrt (SQR(x) + SQR(y));
	r =  c/(1+a*pow((r/b), 2*n));
	H.A[0]=r; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=r;
	return H;

}


/* High-pass filter with n-th order Butterworth response
	H = c / (1 + (1/a)*(R0/R)^2n) = 1-H{LP}
   a is an asymmetry factor,
   b is the cutoff frequency R0 (where the hp filtering begins)
   c is the overall gain 
   d is the order n */

/* Fixed in rev 1.0.0: The above formulation is numerically unacceptable,
	because it leads to a division by zero in the u-v origin. The original
	highpass formula,

		1 / ( 1 + (1/a) (f0 / f)^2n )

	was replaced as of rev. 1.0.0 with

		a F^2n / ( 1 + a F^2n )

	where F is the normalized frequency f/f0.
	Interesting: The division-by-zero highpass formulation is quite common in the literature.
*/


ffilt_mat fft_butterworth_hp (int x, int y, float a, float b, float c, float d)
{
float r;
int n;
ffilt_mat H;

	n = (int)d;
	if (a<0.1) a=0.1;
	r = (sqrt (SQR(x) + SQR(y)))/b;			/* Normalized frequency f/f0 */
	r = pow (r, 2*n);						/* Frequency term w/ order */
	r =   c * a*r / (1+a*r);				/* Final frequency response with asymmetry and gain */
	H.A[0]=r; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=r;
	return H;

}


/* This is experimental and does not work properly. 
	This filter can continuously vary between the Ramachandran kernel 
	and the Shepp-Logan kernel with parameter a.
	The only kludge is that this filter is normalized and needs 
	calibration to image size through parameter d. */
	
	

ffilt_mat fft_first_deriv (int x, int y, float a, float b, float c, float d)
{
float r1, r2;
ffilt_mat H; 

	r1 = sqrt((float) (SQR(x)+SQR(y)));
	r2 = sin (PI*MIN(r1,d)/(2*d))/(PI/2);						/* Shepp-Logan */
	r1 /= d;									/* linear or Ramachandran */
	r1 = (a*r2 + (100.0-a)*r1)/100.0;			/* continuous transition */
//	H.A[0]=0.0; H.A[1]=r; H.A[2]=-r; H.A[3]=0.0;
	H.A[0]=r1; H.A[1]=0.0; H.A[2]=0.0; H.A[3]=r1;

	return H;

}


/* The actual Fourier-based filter function is different from the "naked"
	FFT functions above. Here, we don't need quadrant swapping so we can 
	actually use FFTW3's convenient real-to-complex and complex-to-real
	transformations where the actual filter is a multiplication in the 
	complex Fourier domain.
	Even more, we keep the input data type, i.e. we clamp the filter
	result to fit the original data type (in function fft_store above).

	a, b, c, d are filter coefficients and depend on the filter function.
		c is the DC gain. It was obsoleted in rev 1.0.0 (constant 1.0)
		d must be the image size for the first derivative filter,
			and it must be the order (an integer 1, 2, ...) for the Butterworth filters.

	This function underwent a severe overhaul in rev. 1.0.0 with fixes to the 
	log FFT (homomorphic) and to the window functions. At this point, it appears that
	the FFT on log data actually works...

	For the FFT of log data, care must be taken that the input data is not
	negative (not even zero, in fact). We keep track of the input minimum.
		* If it is 1 or larger, no offset is needed
		* If it is 0 (let's be specific: between 0 and 1), we add an offset of 1
		* If it is negative, we subtract the (negative) minimum and add 1.

	TODO: Optional speedup: Store the window function in a separate image.
	This saves trigonometric functions for the inversion at the end of the FFT.
*/

/* NOTE that this is a single-slice operation, so we need one level higher
	of a function to handle multi-slice images. */

void fourier_filter (image_type* img, int z, gpointer func, int uselog, float a, float b, float c, float d, WINDOW_TYPE window)
{

int xm,ym,x2,ym2;			/* Power-of-two dimensions */
int x,y,i,rgb,idx;
unsigned char preview;
unsigned char* cp;
short* sp;
float* fp;

fftw_real *in;		/* typed pointer to data area */
fftw_complex* out;
fftw_plan fplan;
fftw_plan rplan;
float min, logofs, scalefac;

image_type testimg;
float* t;
ffilt_mat H;
float oldre, oldim;


	fft_filter_func = func;
	cp=img->data; sp=img->data; fp=img->data;
	preview=0; 
	if (z < 0)
	{
		preview=1;
		z=0;
	}

	/* Get memory for transformed image */
	
	xm = img->xmax;
	ym = img->ymax;
	ym2 = ym/2 + (ym & 1);		/* Y-size of the complex out array. For odd y dimensions, add one Fourier row */
	ym2 = ym/2 + 1;				/* Uhh.... no? */
	idx = img->xmax * img->ymax * z;

	dp (1, "FFT: Original image %d by %d; fft output %d by %d\n",xm,ym,xm,ym2);

	in = calloc (xm*ym, sizeof (fftw_real));
	out = calloc ((xm)*ym2, sizeof (fftw_complex));
	if ((in==0) || (out==0)) fatal_memerr();

	if (allocate_image (&testimg, B_FLOAT, xm,ym,1)==-1)
		fatal_memerr();
	t = testimg.data;
	

	/* Fill the new buffer *in with data, converting to float
	   For RGB we'll have to do this three times */
	   
	if (img->imgtype == B_RGB) rgb=3; else rgb=1;
	
	for (i=0; i<rgb; i++)
	{
	
		min = 0.0;				/* Subtract negative offset if present for log conversion */
		switch (img->imgtype)
		{
			case B_CHAR:
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
				   		in[y+ym*x] = cp[x+img->xmax*y+idx];
						if (min > in[x+xm*y]) min = in[x+xm*y];
					}
				break;
			case B_SHORT:
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
				   		in[y+ym*x] = sp[x+img->xmax*y+idx];
						if (min > in[x+xm*y]) min = in[x+xm*y];
					}
				break;
			case B_FLOAT:
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
				   		in[y+ym*x] = fp[x+img->xmax*y+idx];
						if (min > in[x+xm*y]) min = in[x+xm*y];
					}
				break;
			case B_RGB:
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
				   		in[y+ym*x] = cp[3*(x+img->xmax*y+idx)+i];
						if (min > in[x+xm*y]) min = in[x+xm*y];
					}
				break;
		}

		/* If requested, log-transform. Then apply the window function. Note: For log data,
			apply the window function *after* log transform, because this data level is where 
			discontinuities need to be reduced. */

		logofs = 0.0;
		if (min < 0.0)
			logofs = 1.0-min;		/* Make the smallest value 1.0 */
		else if (min < 1.0)
			logofs = 1.0;			/* Add one. This is mostly for integer types */
		if (img->imgtype==B_RGB)
			logofs = 1.0;			/* RGB must have consistent offset. Let's assume that there is always a zero pixel */

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (uselog)
					in[y+ym*x] = SAFELOG(in[y+ym*x]+logofs);

				in[y+ym*x] *= fft_windowfn (window,x,y,img->xmax, img->ymax);
			}

		/* fftw requires us to create a "plan" of the planned transforms.
			In higher debug levels, also output the plan info.
			Note that the r2c and c2r transform imply FORWARD and BACKWARD  */

		fplan = fftw_plan_dft_r2c_2d (xm,ym, in, out, FFTW_ESTIMATE);
		if (debuglevel > 2)
			fftw_print_plan (fplan);

		/* perform the actual transform - call to the library */

		fftw_execute (fplan);

		/* Done - the plan is no longer needed */

		fftw_destroy_plan (fplan);

		/* Apply the filter function */

		if (fft_filter_func)
		{
			for (y=0; y<ym2; y++)
				for (x=0; x<xm/2; x++)
				{
			   		H = fft_filter_func (x,y,a,b,c,d);
				 	
					x2 = xm-1-x;
					oldre = c_re(out[y + ym2*x]); oldim = c_im(out[y + ym2*x]);
					c_re(out[y + ym2*x]) = oldre*H.A[0] + oldim*H.A[1];
				 	c_im(out[y + ym2*x]) = oldre*H.A[2] + oldim*H.A[3];
					oldre = c_re(out[y + ym2*x2]); oldim = c_im(out[y + ym2*x2]);
					c_re(out[y + ym2*x2]) = oldre*H.A[0] + oldim*H.A[1];
					c_im(out[y + ym2*x2]) = oldre*H.A[2] + oldim*H.A[3];
				}
		}


		/* Show the transform AFTER multiplying it with the filter function
			--- but only for debuglevel 2 or higher (display log-magnitude).
			Since the FFT filter makes use of the Hermitian symmetry, we also
			should complete the conjugate half, then swap quadrants. */

		if ((!macrorun) && (!preview) && (debuglevel >= 2))
		{
			for (y=0; y<ym/2; y++)
				for (x=0; x<xm; x++)
				{
					if (x < xm/2)
						i = x + xm/2;
					else
						i = x - xm/2;

					t[i + testimg.xmax*(ym/2-y-1)] = 
					t[i + testimg.xmax*(ym/2+y)] = SAFELOG(sqrt(
															SQR(c_re(out[y + ym2*x]))
															+SQR(c_im(out[y + ym2*x]))
														) +1);
				}
			make_minmax (&testimg);
			secondary_display (&testimg, 0, "FFT Filtered Data");
		}
		
		/* Done - reverse transform */

		rplan = fftw_plan_dft_c2r_2d (xm,ym, out, in, FFTW_ESTIMATE);
		if (debuglevel > 2)
			fftw_print_plan (rplan);

		fftw_execute (rplan);

		fftw_destroy_plan (rplan);

		/* Store back in img.
			If logarithmic, restore the negative
			offset that we subtracted in the beginning. Also,
			normalize the data - a function that fftw doesn't perform
			on its own - and un-window the data */

		scalefac = 1.0/(xm*ym);
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				in[y+ym*x] = in[y+ym*x] * scalefac / fft_windowfn ( window, x,y, img->xmax, img->ymax);
				if (uselog)
					in[y+ym*x] = exp(in[y+ym*x]) - logofs;
		
			}

		fft_store (img, in, xm, ym, z, i);	

	}
	make_minmax (img);
	img->uid = next_uid++;

	/* Free the auxiliary arrays */

	free (in);
	free (out);
	freebuf (&testimg);

}



/* We clone the GUI's "go" function for use from the macro interpreter.
	Note that any change in the filter functions and numbering MUST be reflected
	in macro1.c! */

/* 	TODO: The macro interpreter does not yet allow windowing */

void fftfilt_go (image_type* img, int filt, float a, float b, float c, float d, WINDOW_TYPE wind)
{
int z, size;

	switch (filt)
	{
		case 1: 
			fft_filter_func = &fft_homomorphic_hp;
			break;
		case 2: 
			fft_filter_func = &fft_gaussian_lp;
			break;
		case 3: 
			fft_filter_func = &fft_butterworth_lp;
			break;
		case 4: 
			fft_filter_func = &fft_butterworth_hp;
			break;
		case 5: 
			fft_filter_func = &fft_first_deriv;
			size = MAX(img->xmax, img->ymax);
			d = size/2;							/* Overwrite parameter d for first derivative filter */
			break;
		case 6: 
			fft_filter_func = &fft_wiener;
			break;
		case 7: 
			fft_filter_func = &fft_oppenheim_hp;
			break;
		default:
			fft_filter_func = &neutral_fft_filter;
	}

	for (z=0; z<img->zmax; z++)
	{
		dp (1, "Fourier fiter: Now processing slice %d\n", z);
		fourier_filter (img, z, fft_filter_func, 0, a,b,c,d, wind);
	}
}


void fftfilt_log_go (image_type* img, int filt, float a, float b, float c, float d, WINDOW_TYPE wind)
{
int z, size;

	switch (filt)
	{
		case 1: 
			fft_filter_func = &fft_homomorphic_hp;
			break;
		case 2: 
			fft_filter_func = &fft_gaussian_lp;
			break;
		case 3: 
			fft_filter_func = &fft_butterworth_lp;
			break;
		case 4: 
			fft_filter_func = &fft_butterworth_hp;
			break;
		case 5: 
			fft_filter_func = &fft_first_deriv;
			size = MAX(img->xmax, img->ymax);
			d = size/2;							/* Overwrite parameter d for first derivative filter */
			break;
		case 6: 
			fft_filter_func = &fft_wiener;
			break;
		default:
			fft_filter_func = &neutral_fft_filter;
	}

	for (z=0; z<img->zmax; z++)
	{
		dp (1, "Fourier fiter: Now processing log slice %d\n", z);
		fourier_filter (img, z, fft_filter_func, 1, a,b,c,d, wind);
	}
}





/*************************************************************

    	The FFT filter GUI 
	
**************************************************************/

GtkWidget *fftfilt_wnd;
GtkAdjustment *fftfilt_a_adj, *fftfilt_b_adj, *fftfilt_c_adj, *fftfilt_d_adj, *fftfilt_imgsize_adj;
GtkWidget *fftfilt_a, *fftfilt_b, *fftfilt_c, *fftfilt_d, *fftfilt_imgsize;
GtkWidget *fftfilt_label1, *fftfilt_label2, *fftfilt_label3, *fftfilt_label4, *fftfilt_label5;
GtkWidget *fftfilt_draw;

#define FFTFILT_GRAPH_MAXPTS	128

int fftfilt_function;
WINDOW_TYPE fftfilt_window;
int fftfilt_uselog;
image_type fftfilt_preview_img;
image_type* fftfilt_img;
float fftfilt_xdata[FFTFILT_GRAPH_MAXPTS], fftfilt_ydata[FFTFILT_GRAPH_MAXPTS];


/* This gets called whenever a parameter changes. It simply
   redisplays the H(f) graph. Note that it relies on skeleton functions
   from graphdsp.c, and we need to manage some of the data structures externally. */

void fftfilt_compute (GtkWidget* widget, gpointer user_data)
{ 
int i,imax;
float deltai;
float a,b,c, ord;
ffilt_mat H;

	imax = MAX(fftfilt_preview_img.xmax, fftfilt_preview_img.ymax);
	deltai = imax/(4.0*FFTFILT_GRAPH_MAXPTS);
	
	a = fftfilt_a_adj->value;
	b = fftfilt_b_adj->value;
	c = fftfilt_c_adj->value;
	ord = fftfilt_d_adj->value;

	if (fftfilt_function==5)  c=FFTFILT_GRAPH_MAXPTS;
	if ((fftfilt_function==3) || (fftfilt_function==4))
	{
		c = ord;		/* Fourth coeff is the order for Butterworth filters */
	}
	
	for (i=0; i<FFTFILT_GRAPH_MAXPTS; i++)
	{
		fftfilt_xdata[i] = deltai * i;
		H = fft_filter_func (deltai*i, 0, a,b,1.0,c);
		fftfilt_ydata[i] = sqrt(SQR(H.A[0])+SQR(H.A[1])+SQR(H.A[2])+SQR(H.A[3]))/SQ2;
	}

	graph_labels ("f (1/px)", "H", "Frequency Response", 0);
	Frame (	fftfilt_draw, fftfilt_xdata, fftfilt_ydata, FFTFILT_GRAPH_MAXPTS, 1);
	DrawData (	fftfilt_draw, fftfilt_xdata, fftfilt_ydata, FFTFILT_GRAPH_MAXPTS, 1);

}


/* Context-dependent controls. We have a maximum of 4 coefficients,
	a, b, c, d, which are used as follows:

	Filter					a			b			c			d			Sliders
	No filter				*			*			*			*			4
	Gaussian LP				Sigma		Frq.ofs		-			-			2
	Butterworth LP			Asymm		Cutoff		-			Order		2+1
	Wiener filter			Sigma		Cutoff		Denom ofs.	-			3
	Homomorphic HP			Steepness	Cutoff		-			-			2
	Butterworth HP			Asymm		Cutoff		-			Order		2+1
	Oppenheim HP			DC atten.	Steepness	-			-			2
	First derivative		SH-"bending"	-		-			-			1

	Notes:
		* indicates the slider is available, but not evaluated
		- indicates that the slider is hidden
		Order (Butterworth): Integer
		Asymmetry (Butterworth): Should default to 1
		Coefficients become visible as needed (a, ab+b, a+b+c, ...) with
		the exception of the order, which is an integer, and its slider configured
		to have zero decimals.
		As of rev. 1.0.0, the ridiculous gain slider has been removed. Gain is maintained,
		however, so that e.g., a frequency response can be obtained in percent.
*/


void fftfilt_set_menu (GtkWidget *widget, gpointer user_data)
{

	fftfilt_function = *(int*) user_data;
	
	switch (fftfilt_function)
	{
		case 1: 
			fft_filter_func = &fft_homomorphic_hp;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "Steepness s");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Cutoff r(0)");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Coefficient c");
			break;
		case 2: 
			fft_filter_func = &fft_gaussian_lp;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "Steepness sigma");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Cutoff r(0)");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Coefficient c");
			break;
		case 3: 
			fft_filter_func = &fft_butterworth_lp;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "Asymmetry");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Cutoff r(0)");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Coefficient c");
			gtk_label_set_text (GTK_LABEL(fftfilt_label5), "Order n");
			break;
		case 4: 
			fft_filter_func = &fft_butterworth_hp;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "Asymmetry");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Cutoff r(0)");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Coefficient c");
			gtk_label_set_text (GTK_LABEL(fftfilt_label5), "Order n");
			break;
		case 5: 
			fft_filter_func = &fft_first_deriv;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "Shepp-Logan-ness (\%)");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Coefficient b");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Coefficient c");
			break;
		case 6: 
			fft_filter_func = &fft_wiener;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "Steepness sigma");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Cutoff r(0)");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Wiener offset");
			break;
		case 7: 
			fft_filter_func = &fft_oppenheim_hp;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "DC attenuation (\%)");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Steepness s");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Coefficient c");
			break;
		default:
			fft_filter_func = &neutral_fft_filter;
			gtk_label_set_text (GTK_LABEL(fftfilt_label2), "Coefficient a");
			gtk_label_set_text (GTK_LABEL(fftfilt_label3), "Coefficient b");
			gtk_label_set_text (GTK_LABEL(fftfilt_label4), "Coefficient c");
	}

	switch (fftfilt_function)		/* Make individual filter coefficient sliders visible/invisible */
	{
		case 1:					/* Gaussian, Homomorphic, Oppenheim: */
		case 2:					/* Two coefficients visible */
		case 7:
			gtk_widget_show (fftfilt_label2); gtk_widget_show (fftfilt_a);
			gtk_widget_show (fftfilt_label3); gtk_widget_show (fftfilt_b);
			gtk_widget_hide (fftfilt_label4); gtk_widget_hide (fftfilt_c);
			gtk_widget_hide (fftfilt_label5); gtk_widget_hide (fftfilt_d);
			break;

		case 3:					/* Butterworth filters */
		case 4:					/* Two coefficients + order */
			gtk_widget_show (fftfilt_label2); gtk_widget_show (fftfilt_a);
			gtk_widget_show (fftfilt_label3); gtk_widget_show (fftfilt_b);
			gtk_widget_hide (fftfilt_label4); gtk_widget_hide (fftfilt_c);
			gtk_widget_show (fftfilt_label5); gtk_widget_show (fftfilt_d);
			break;

		case 5:					/* One coefficient visible */
			gtk_widget_show (fftfilt_label2); gtk_widget_show (fftfilt_a);
			gtk_widget_hide (fftfilt_label3); gtk_widget_hide (fftfilt_b);
			gtk_widget_hide (fftfilt_label4); gtk_widget_hide (fftfilt_c);
			gtk_widget_hide (fftfilt_label5); gtk_widget_hide (fftfilt_d);
			break;
		case 6:					/* Wiener filter: Three coefficients visible */
			gtk_widget_show (fftfilt_label2); gtk_widget_show (fftfilt_a);
			gtk_widget_show (fftfilt_label3); gtk_widget_show (fftfilt_b);
			gtk_widget_show (fftfilt_label4); gtk_widget_show (fftfilt_c);
			gtk_widget_hide (fftfilt_label5); gtk_widget_hide (fftfilt_d);
			break;
		default:			/* all coefficients visible */
			gtk_widget_show (fftfilt_label2); gtk_widget_show (fftfilt_a);
			gtk_widget_show (fftfilt_label3); gtk_widget_show (fftfilt_b);
			gtk_widget_show (fftfilt_label4); gtk_widget_show (fftfilt_c);
			gtk_widget_show (fftfilt_label5); gtk_widget_show (fftfilt_d);
			break;
	}

	/* Redisplay the filter function */
	fftfilt_compute (NULL, NULL);
	
}


void fftfilt_set_wmenu (GtkWidget *widget, gpointer user_data)
{

	fftfilt_window = *(int*) user_data;		/* NOTE: Must match the WINDOW_TYPE enum */
	dp (2, "FFT filter: Switched filter function to ");
	switch (fftfilt_window)
	{
		case W_NONE:					/* Opt 0 */
			dp (2, "none\n");
			break;
		case W_WELCH:					/* Opt 1 */
			dp (2, "Welch\n");
			break;
		case W_PARZEN:					/* Opt 2 */
			dp (2, "Parzen\n");
			break;
		case W_HANN:					/* Opt 3 */
			dp (2, "Hann\n");
			break;
		case W_HAMMING:					/* Opt 4 */
			dp (2, "Hamming\n");
			break;
		case W_GAUSS:					/* Opt 5 */
			dp (2, "Gaussian\n");
			break;
		default:
			dp (2, "Unknown/undefined\n");
	}


}



void on_fftfilt_checkbutton1_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{

	fftfilt_uselog = !fftfilt_uselog;
}

void on_fftfilt_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	freebuf (&fftfilt_preview_img);
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_fftfilt_preview_clicked  (GtkButton *button, gpointer user_data)
{
float a,b,c, ord;
int e; long len;


	a = fftfilt_a_adj->value;
	b = fftfilt_b_adj->value;
	c = fftfilt_c_adj->value;
	ord = fftfilt_d_adj->value;
	if (fftfilt_function==5) c=MAX (fftfilt_img->xmax, fftfilt_img->ymax)/2;
	if ((fftfilt_function==3) || (fftfilt_function==4))
	{
		c = ord;		/* Fourth coeff is the order for Butterworth filters */
	}


	/* Use the main filter function to filter the current slice,
	   then display it. This is not the most efficient way (we might
	   forward-transform our buffer just once, then reuse it), but
	   quicker to write and less prone to errors.

	   By passing a slice of (-1) to fourier_filter(), we indicate that it is a preview,
	   and thus prevent the filtered FFT data from being displayed.

	   It is important to retain the data type in the preview mode, because
	   we need to preview the filter result with all its artifacts */


	e = DATALEN (fftfilt_img->imgtype);
	len = fftfilt_img->xmax * fftfilt_img->ymax * e;
	memcpy (fftfilt_preview_img.data, fftfilt_img->data, len);
	fourier_filter (&fftfilt_preview_img, -1, fft_filter_func, fftfilt_uselog, a,b,1.0,c, fftfilt_window);

	secondary_display (&fftfilt_preview_img, 0, "Filter Preview");

}


/* Alternative preview: View the window function (for debugging only) */

void on_fftwind_preview_clicked  (GtkButton *button, gpointer user_data)
{
int x,y, xm,ym;
image_type preview;

	xm = mainimg.xmax; ym = mainimg.ymax;
	if (allocate_image (&preview, B_FLOAT, xm, ym, 1)==-1)
		fatal_memerr();

	for (y=0; y<ym; y++)
		for (x=0; x<xm; x++)
			((float *)preview.data)[x + xm*y] = fft_windowfn ( fftfilt_window, x, y, xm, ym);

	make_minmax (&preview);
	secondary_display (&preview, 0, "Window Preview");
}



/* After all this trouble, the actual "go" function is easy... */

void on_fftfilt_ok_clicked  (GtkButton *button, gpointer user_data)
{
float a,b,c, ord;
int z;

	if (fftfilt_img == &mainimg) undo_nexus (fftfilt_img);

	a = fftfilt_a_adj->value;
	b = fftfilt_b_adj->value;
	c = fftfilt_c_adj->value;
	ord = fftfilt_d_adj->value;
	if (fftfilt_function==5) c=MAX (fftfilt_img->xmax, fftfilt_img->ymax)/2;
	if ((fftfilt_function==3) || (fftfilt_function==4))
	{
		c = ord;		/* Fourth coeff is the order for Butterworth filters */
	}

	for (z=0; z<fftfilt_img->zmax; z++)
	{
		dp (1, "Fourier filter: Now processing slice %d\n", z);

		fourier_filter (fftfilt_img, z, fft_filter_func, fftfilt_uselog, a,b,1.0,c, fftfilt_window);
	}
	if (fftfilt_img == &mainimg) 
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
	}
	
	freebuf (&fftfilt_preview_img);
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


/* The alternative to the OK button. This callback prepares the coefficients and
	returns them to the external function.

	I admit that this structure seriously violates the concept of the DSOs, and
	the mess of function pointers is a terrible price to pay for... what, actually?
	For the convenience that any changes or added functions here translate directly
	into the options in the generator module.

	TODO: If we have this ... ahem... convenience anyway, can't we add the
	windowing functions to the generator options?
*/


void on_fftfilt_ext_clicked  (GtkButton *button, gpointer user_data)
{
float a,b,c, ord;
int size;
void (*ext_function_proto)  (gpointer func, int size, float a, float b, float c, float d);


	ext_function_proto = user_data;
	a = fftfilt_a_adj->value;
	b = fftfilt_b_adj->value;
	c = fftfilt_c_adj->value;
	ord = fftfilt_d_adj->value;
	size = fftfilt_imgsize_adj->value;
	if (fftfilt_function==5) c=size;
	if ((fftfilt_function==3) || (fftfilt_function==4))
	{
		c = ord;		/* Fourth coeff is the order for Butterworth filters */
	}

	ext_function_proto (fft_filter_func, size, a, b, 100.0, c);
	freebuf (&fftfilt_preview_img);
	gtk_widget_destroy (GTK_WIDGET(fftfilt_wnd));
}

void on_fftfilt_ext2_clicked  (GtkButton *button, gpointer user_data)
{
int a, size;
void (*ext_function_proto)  (gpointer func, int size, float a, float b, float c, float d);


	ext_function_proto = user_data;		/* Same callback function in generator.c */
	a = fftfilt_window;					/* except that the MTF pointer is NULL */
	size = fftfilt_imgsize_adj->value;	/* Size is the only slider actually used */

	ext_function_proto (NULL, size, a, 0.0, 1.0, 0.0);
	freebuf (&fftfilt_preview_img);
	gtk_widget_destroy (GTK_WIDGET(fftfilt_wnd));
}




/* The main GUI function. I want to be able to use this function from the generator module,
	where we'd have a different function assigned to the "OK" button. Pass it as function pointer.
	If it is NULL, we use the default for FFT. */


void create_fftfilt_wnd (image_type* img, void *ext_funct)
{
  GtkWidget *vbox1;
  GtkWidget *table1;
  GtkWidget *fftfilt_optionmenu1;
  GtkWidget *fftfilt_optionmenu1_menu;
  GtkWidget *fftfilt_optionmenu2;
  GtkWidget *fftfilt_optionmenu2_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *fftfilt_label6, *fftfilt_label7, *fftfilt_label8;
  GtkWidget *fftfilt_uselog;
  GtkWidget *hbox1, *hsep1, *hsep2;
  GtkWidget *fftfilt_okbutton;
  GtkWidget *fftfilt_cancelbutton;
  GtkWidget *fftfilt_previewbutton;
  GtkTooltips *tt_ft_gauss, *tt_ft_btwlp, *tt_ft_wft, *tt_ft_hhp, *tt_ft_bhp, *tt_ft_ohp, *tt_ft_der;
  GtkTooltips *tt_ft_noop, *tt_ft_am, *tt_ft_wnd, *tt_ft_prev, *tt_ft_xt, *tt_ft_go;
  int y;


	fftfilt_img = NULL;			/* Catch problems with undefined pointers */
	if (img)
	{
		fftfilt_img = img;
		if (allocate_image (&fftfilt_preview_img, img->imgtype, img->xmax, img->ymax, 1)==-1)
			fatal_memerr();
	}
	else		/* We still have to define this image for compatibility, let's use the default size */
	{
		if (allocate_image (&fftfilt_preview_img, B_FLOAT, 512, 512, 1)==-1)
			fatal_memerr();
	}


	fftfilt_uselog = 0; fftfilt_window=0;
	fft_filter_func = &neutral_fft_filter;
	fftfilt_function = 0;

	fftfilt_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (fftfilt_wnd), "fftfilt_wnd", fftfilt_wnd);
	gtk_window_set_title (GTK_WINDOW (fftfilt_wnd), "Frequency Domain Filters");

	vbox1 = vbox_in_window (fftfilt_wnd,"vbox1"); 
	table1 = MakeTable (fftfilt_wnd, vbox1, "table1", 2,7);


	y=0;
	fftfilt_label1 = MakeLabel (fftfilt_wnd, table1, "Filter type", "label1", 0,y);
	fftfilt_optionmenu1 = Make_Optionmenu (fftfilt_wnd, table1,"fftfilt_menu",1,y);
	gtk_container_set_border_width (GTK_CONTAINER (fftfilt_optionmenu1), 1);
	fftfilt_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "No Filter Selected", (gpointer) &menuopt0);
	tt_ft_noop = MakeTooltip (glade_menuitem, "Pass-through for debugging purposes.\nFilter does not perform any action.");
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "Gaussian LP", (gpointer) &menuopt2);
	tt_ft_gauss = MakeTooltip (glade_menuitem, "Lowpass with Gaussian dropoff and\nadjustable horizontal shoulder");
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "Butterworth LP", (gpointer) &menuopt3);
	tt_ft_btwlp = MakeTooltip (glade_menuitem, "Butterworth Lowpass of arbitrary order\nand asymmetric dropoff.");
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "Wiener filter", (gpointer) &menuopt6);
	tt_ft_wft = MakeTooltip (glade_menuitem, "Wiener filter (image-restoring highpass)");
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "Homomorphic HP", (gpointer) &menuopt1);
	tt_ft_hhp = MakeTooltip (glade_menuitem, "Homomorphic highpass often used for background removal");
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "Butterworth HP", (gpointer) &menuopt4);
	tt_ft_bhp = MakeTooltip (glade_menuitem, "Butterworth highpass of arbitrary order\nand asymmetric dropoff.");
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "Oppenheim HP", (gpointer) &menuopt7);
	tt_ft_ohp = MakeTooltip (glade_menuitem, "Oppenheim highpass.");
	glade_menuitem = Option_menuitem (fftfilt_optionmenu1_menu, GTK_SIGNAL_FUNC(fftfilt_set_menu), "First derivative", (gpointer) &menuopt5);
	tt_ft_der = MakeTooltip (glade_menuitem, "First-derivative via FFT with\nadjustable high-frequency attenuation");
	gtk_option_menu_set_menu (GTK_OPTION_MENU (fftfilt_optionmenu1), fftfilt_optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (fftfilt_optionmenu1), 0);

	if (ext_funct)
	{
		y++;
		fftfilt_label8 = MakeLabel (fftfilt_wnd, table1, "Image size", "mkfftmask_label8", 0,y);
		fftfilt_imgsize_adj = GTK_ADJUSTMENT (gtk_adjustment_new (512, 128, 2048, 1, 128, 0));
		fftfilt_imgsize = MakeHScale (fftfilt_imgsize_adj, fftfilt_wnd, table1,"mkfftmask_imgsize", 1,y);
		gtk_scale_set_digits (GTK_SCALE (fftfilt_imgsize), 0);
	}

	y++;
	fftfilt_label2 = MakeLabel (fftfilt_wnd, table1, "Parameter a", "fftfilt_label2", 0,y);
	fftfilt_a_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0, 100, 0.1, 1, 0));
	fftfilt_a = MakeHScale (fftfilt_a_adj, fftfilt_wnd, table1,"fftfilt_a", 1,y);

	y++;
	fftfilt_label3 = MakeLabel (fftfilt_wnd, table1, "Parameter b", "fftfilt_label3", 0,y);
	fftfilt_b_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, 0, 100, 0.1, 1, 0));
	fftfilt_b = MakeHScale (fftfilt_b_adj, fftfilt_wnd, table1,"fftfilt_b", 1,y);

	y++;			/* Usually gain -- give it a reasonable default value of unity */
	fftfilt_label4 = MakeLabel (fftfilt_wnd, table1, "Parameter c", "fftfilt_label4", 0,y);
	fftfilt_c_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0, 100, 0.1, 1, 0));
	fftfilt_c = MakeHScale (fftfilt_c_adj, fftfilt_wnd, table1,"fftfilt_c", 1,y);

	y++;
	fftfilt_label5 = MakeLabel (fftfilt_wnd, table1, "Order n", "fftfilt_label5", 0,y);
	fftfilt_d_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 1, 10, 1, 1, 0));
	fftfilt_d = MakeHScale (fftfilt_d_adj, fftfilt_wnd, table1,"fftfilt_d", 1,y);
	gtk_scale_set_digits (GTK_SCALE (fftfilt_d), 0);		/* Order is integer */

	y++;
	hsep1 = MakeHsep (fftfilt_wnd, table1, "hsep1", 0,y); hsep1 = MakeHsep (fftfilt_wnd, table1, "hsep1", 1,y);

	y++;
	fftfilt_label6 = MakeLabel (fftfilt_wnd, table1, "Additive/Multiplicative", "fftfilt_label6", 0,y);
	fftfilt_uselog = Checkbutton_in_table (fftfilt_wnd, table1, "FFT on log data","fftfilt_uselog",1,y, on_fftfilt_checkbutton1_toggled);
	tt_ft_am = MakeTooltip (fftfilt_uselog, "EXPERIMENTAL: When checked, filters log data\nwhich interprets the removed components as multiplicative (e.g., inhomogeneous illumination)");

	y++;
	fftfilt_label7 = MakeLabel (fftfilt_wnd, table1, "Window function", "fftfilt_label7", 0,y);
	fftfilt_optionmenu2 = Make_Optionmenu (fftfilt_wnd, table1,"fftfilt_menu2",1,y);
	gtk_container_set_border_width (GTK_CONTAINER (fftfilt_optionmenu2), 1);
	fftfilt_optionmenu2_menu = gtk_menu_new ();
	tt_ft_wnd = MakeTooltip (fftfilt_optionmenu2, "Select windowing of image data");

	/* The menuoptx pointers must match the WINDOW_TYPE enum */
	glade_menuitem = Option_menuitem (fftfilt_optionmenu2_menu, GTK_SIGNAL_FUNC(fftfilt_set_wmenu), "None", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (fftfilt_optionmenu2_menu, GTK_SIGNAL_FUNC(fftfilt_set_wmenu), "Hamming", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (fftfilt_optionmenu2_menu, GTK_SIGNAL_FUNC(fftfilt_set_wmenu), "Hann", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (fftfilt_optionmenu2_menu, GTK_SIGNAL_FUNC(fftfilt_set_wmenu), "Gaussian", (gpointer) &menuopt5);
	glade_menuitem = Option_menuitem (fftfilt_optionmenu2_menu, GTK_SIGNAL_FUNC(fftfilt_set_wmenu), "Welch", (gpointer) &menuopt1);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (fftfilt_optionmenu2), fftfilt_optionmenu2_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (fftfilt_optionmenu2), 0);

	/* Next is a drawing area in which we can plot the frequency response */

	hsep1 = MakeHsep (fftfilt_wnd, vbox1, "hsep1", -1,-1);
	fftfilt_draw = gtk_drawing_area_new ();
	gtk_widget_ref (fftfilt_draw);
	gtk_object_set_data_full (GTK_OBJECT (fftfilt_wnd), "fftfilt_draw", fftfilt_draw,
                            (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (fftfilt_draw);
	gtk_box_pack_start (GTK_BOX (vbox1), fftfilt_draw, TRUE, TRUE, 4);
	gtk_widget_set_usize (fftfilt_draw, 200, 150);
	hsep2 = MakeHsep (fftfilt_wnd, vbox1, "hsep2", -1,-1);

	hbox1 = gtk_hbox_new (FALSE, 0);
	gtk_widget_ref (hbox1);
	gtk_object_set_data_full (GTK_OBJECT (fftfilt_wnd), "hbox1", hbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox1);
	gtk_box_pack_start (GTK_BOX (vbox1), hbox1, TRUE, TRUE, 0);

	if (ext_funct)
	{
		fftfilt_okbutton = Button_in_table (fftfilt_wnd, hbox1, "Filter MTF", "fftfilt_okbutton",-1,-1,NULL);
		tt_ft_go = MakeTooltip (fftfilt_okbutton, "Generates 2D MTF of the selected filter");
		gtk_signal_connect (GTK_OBJECT (fftfilt_okbutton), "clicked", GTK_SIGNAL_FUNC (on_fftfilt_ext_clicked), ext_funct);
		fftfilt_previewbutton = Button_in_table (fftfilt_wnd, hbox1, "Window", "fftfilt_previewbutton",-1,-1,NULL);
		tt_ft_prev = MakeTooltip (fftfilt_previewbutton, "Generates 2D window function (if selected)");
		gtk_signal_connect (GTK_OBJECT (fftfilt_previewbutton), "clicked", GTK_SIGNAL_FUNC (on_fftfilt_ext2_clicked), ext_funct);
	}
	else
	{
		fftfilt_okbutton = Button_in_table (fftfilt_wnd, hbox1, "OK", "fftfilt_okbutton",-1,-1,NULL);
		tt_ft_go = MakeTooltip (fftfilt_okbutton, "Filters the main image with the selected filter function");
		gtk_signal_connect (GTK_OBJECT (fftfilt_okbutton), "clicked", GTK_SIGNAL_FUNC (on_fftfilt_ok_clicked), fftfilt_wnd);
		fftfilt_previewbutton = Button_in_table (fftfilt_wnd, hbox1, "Preview", "fftfilt_previewbutton",-1,-1,NULL);
		tt_ft_prev = MakeTooltip (fftfilt_previewbutton, "Allows to preview the filtered data in secondary window");
		gtk_signal_connect (GTK_OBJECT (fftfilt_previewbutton), "clicked", GTK_SIGNAL_FUNC (on_fftfilt_preview_clicked), fftfilt_wnd);
	}

	fftfilt_cancelbutton = Button_in_table (fftfilt_wnd, hbox1, "CANCEL", "fftfilt_cancelbutton",-1,-1,NULL);
	tt_ft_xt = MakeTooltip (fftfilt_okbutton, "Exit without applying the filter");

	gtk_signal_connect (GTK_OBJECT (fftfilt_a_adj), "value_changed", GTK_SIGNAL_FUNC (fftfilt_compute), &menuopt0);
	gtk_signal_connect (GTK_OBJECT (fftfilt_b_adj), "value_changed", GTK_SIGNAL_FUNC (fftfilt_compute), &menuopt1);
	gtk_signal_connect (GTK_OBJECT (fftfilt_c_adj), "value_changed", GTK_SIGNAL_FUNC (fftfilt_compute), &menuopt2);
	gtk_signal_connect (GTK_OBJECT (fftfilt_d_adj), "value_changed", GTK_SIGNAL_FUNC (fftfilt_compute), &menuopt3);
	gtk_signal_connect (GTK_OBJECT (fftfilt_draw), "expose_event", GTK_SIGNAL_FUNC (fftfilt_compute), NULL);
	gtk_signal_connect (GTK_OBJECT (fftfilt_cancelbutton), "clicked", GTK_SIGNAL_FUNC (on_fftfilt_cancel_clicked), fftfilt_wnd);
	gtk_signal_connect (GTK_OBJECT (fftfilt_wnd), "destroy", GTK_SIGNAL_FUNC (on_fftfilt_cancel_clicked), fftfilt_wnd);

	/* The OK button may be connected to an external function (if the pointer is not NULL).
		In that case, we also hide the "preview" button. */




	gtk_widget_show (fftfilt_wnd); 

}




void fourierfilt_main (image_type* img)
{
	create_fftfilt_wnd (img, NULL);
}






/************************************************************

One-dimensional FFT filter as utility
function wherever a filter is needed.
The 1D filter uses a similar interface as the 2D filter,
meaning, we give it the requested filter transfer function
as callback in (global) fft_filter_func.


***************************************************************/


void fft_1d_filter (float *databuf, int npoints, gpointer func, int flags, float a, float b, float c, float d)
{
fftw_complex *out;		/* typed pointer to data area */
fftw_real *in;			/* typed pointer to complex input */
int i,n,k;
float oldre, oldim;
ffilt_mat H;
char apodized;

fftw_plan fplan, bplan;


	/* Prepare input buffer, any other buffers, and the plan */

	n = npoints; if (n | 1) n++;		/* Make sure we have an even number, pad with zero if necessary */
	in = (fftw_real*) fftw_malloc(sizeof(fftw_complex) * n);
	out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * n);
	fplan = fftw_plan_dft_r2c_1d(n, in, out, FFTW_ESTIMATE);
	bplan = fftw_plan_dft_c2r_1d(n, out, in, FFTW_ESTIMATE);
	
	for (i=0; i<npoints; i++) in[i]=databuf[i];
	if (n>npoints) for (i=npoints; i<n; i++) in[i]=0.0; 

	/* Apply an apodization window if requested */

	apodized=0;
	k = MAX(n/10,64);					/* Number of points to use for cosine tapering */
	if ((k>=10) && (flags & 0x01))		/* Cosine-taper window requested and reasonable? */
	{
		for (i=0; i<k; i++)
		{
			in[i] *= 0.55 - 0.45*cos(PI*i/k);
			in[n-i-1] *= 0.55 - 0.45*cos(PI*i/k);
		}
		apodized=1;	/* Note: Don't overwrite k up to the end. */
	}

	/* Time to do the forward transform */

	fftw_execute(fplan);

	/* Now apply the filter, normalize at the same time */

	if (func)
	{
		fft_filter_func = func;
		for (i=0; i<n/2; i++)
		{
			H = fft_filter_func (i,0,a,b,c,d);

			oldre = c_re(out[i]); oldim = c_im(out[i]);
			c_re(out[i]) = oldre*H.A[0] + oldim*H.A[1];
			c_im(out[i]) = oldre*H.A[2] + oldim*H.A[3];
			if (i>0)				/* Mirrored negative frequencies */
			{
				oldre = c_re(out[n-i]); oldim = c_im(out[n-i]);
				c_re(out[n-i]) = oldre*H.A[0] + oldim*H.A[1];
				c_im(out[n-i]) = oldre*H.A[2] + oldim*H.A[3];
			}
		}
	}

	fftw_execute (bplan);
	for (i=0; i<npoints; i++) in[i]=in[i]/n;		/* Normalization */

	/* De-apodize if necessary, then stuff back in databuffer */

	if (apodized)
	{
		for (i=0; i<k; i++)
		{
			in[i] /= 0.55 - 0.45*cos(PI*i/k);
			in[n-i-1] /= 0.55 - 0.45*cos(PI*i/k);
		}
	}

	for (i=0; i<npoints; i++) databuf[i]=in[i];			/* double->float loss of precision acceptable */

	fftw_destroy_plan(fplan);
	fftw_destroy_plan(bplan);
	fftw_free(in); fftw_free(out);

}










