

/*
 *  interpolate.c - Interpolation, as the name implies.
 * Most importantly, this module contains the spline-based
 * interpolation algorithms
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.

		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>


#include "defs.h"
#include "functions.h"
#include "imgbuf.h"
#include "interpolate.h"



/* TODO: Linear and cubic belong here, too */







/***************************************************************

	Spline-based interpolation

 *	This section is based on the following three papers:
 *		[1]	M. Unser,
 *			"Splines: A Perfect Fit for Signal and Image Processing,"
 *			IEEE Signal Processing Magazine, vol. 16, no. 6, pp. 22-38,
 *			November 1999.
 *		[2]	M. Unser, A. Aldroubi and M. Eden,
 *			"B-Spline Signal Processing: Part I--Theory,"
 *			IEEE Transactions on Signal Processing, vol. 41, no. 2, pp. 821-832,
 *			February 1993.
 *		[3]	M. Unser, A. Aldroubi and M. Eden,
 *			"B-Spline Signal Processing: Part II--Efficient Design and Applications,"
 *			IEEE Transactions on Signal Processing, vol. 41, no. 2, pp. 834-848,
 *			February 1993.
 *
 *	Credits:  Philippe Thevenaz, EPFL/STI/IOA/BIG (Lausanne, CH)
 *
***************************************************************/

image_type splcoeff;


void compute_weights (double x, double y, int* xi, int* yi, double* xw, double* yw, int width, int height, int SplDeg)
{
int i,j,k;
int	Width2, Height2;
double	w, w2, w4, t, t0, t1;

	Width2 = 2 * width - 2;
	Height2 = 2 * height - 2;

	/* compute the interpolation indices */

	if (SplDeg & 1) 	/* Degree 3 or 5 */
	{
		i = (int)floor(x) - SplDeg / 2;
		j = (int)floor(y) - SplDeg / 2;
		for (k = 0; k <= SplDeg; k++)
		{
			xi[k] = i++;
			yi[k] = j++;
		}
	}
	else 	/* degree 2 or 4 */
	{
		i = (int)floor(x + 0.5) - SplDeg / 2;
		j = (int)floor(y + 0.5) - SplDeg / 2;
		for (k = 0; k <= SplDeg; k++)
		{
			xi[k] = i++;
			yi[k] = j++;
		}
	}

	/* compute the interpolation weights */

	switch (SplDeg)
	{
		case 2:
			/* x */
			w = x - (double)xi[1];
			xw[1] = 3.0 / 4.0 - w * w;
			xw[2] = (1.0 / 2.0) * (w - xw[1] + 1.0);
			xw[0] = 1.0 - xw[1] - xw[2];
			/* y */
			w = y - (double)yi[1];
			yw[1] = 3.0 / 4.0 - w * w;
			yw[2] = (1.0 / 2.0) * (w - yw[1] + 1.0);
			yw[0] = 1.0 - yw[1] - yw[2];
			break;
		case 3:
			/* x */
			w = x - (double)xi[1];
			xw[3] = (1.0 / 6.0) * w * w * w;
			xw[0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - xw[3];
			xw[2] = w + xw[0] - 2.0 * xw[3];
			xw[1] = 1.0 - xw[0] - xw[2] - xw[3];
			/* y */
			w = y - (double)yi[1];
			yw[3] = (1.0 / 6.0) * w * w * w;
			yw[0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - yw[3];
			yw[2] = w + yw[0] - 2.0 * yw[3];
			yw[1] = 1.0 - yw[0] - yw[2] - yw[3];
			break;
		case 4:
			/* x */
			w = x - (double)xi[2];
			w2 = w * w;
			t = (1.0 / 6.0) * w2;
			xw[0] = 1.0 / 2.0 - w;
			xw[0] *= xw[0];
			xw[0] *= (1.0 / 24.0) * xw[0];
			t0 = w * (t - 11.0 / 24.0);
			t1 = 19.0 / 96.0 + w2 * (1.0 / 4.0 - t);
			xw[1] = t1 + t0;
			xw[3] = t1 - t0;
			xw[4] = xw[0] + t0 + (1.0 / 2.0) * w;
			xw[2] = 1.0 - xw[0] - xw[1] - xw[3] - xw[4];
			/* y */
			w = y - (double)yi[2];
			w2 = w * w;
			t = (1.0 / 6.0) * w2;
			yw[0] = 1.0 / 2.0 - w;
			yw[0] *= yw[0];
			yw[0] *= (1.0 / 24.0) * yw[0];
			t0 = w * (t - 11.0 / 24.0);
			t1 = 19.0 / 96.0 + w2 * (1.0 / 4.0 - t);
			yw[1] = t1 + t0;
			yw[3] = t1 - t0;
			yw[4] = yw[0] + t0 + (1.0 / 2.0) * w;
			yw[2] = 1.0 - yw[0] - yw[1] - yw[3] - yw[4];
			break;
		case 5:
			/* x */
			w = x - (double)xi[2];
			w2 = w * w;
			xw[5] = (1.0 / 120.0) * w * w2 * w2;
			w2 -= w;
			w4 = w2 * w2;
			w -= 1.0 / 2.0;
			t = w2 * (w2 - 3.0);
			xw[0] = (1.0 / 24.0) * (1.0 / 5.0 + w2 + w4) - xw[5];
			t0 = (1.0 / 24.0) * (w2 * (w2 - 5.0) + 46.0 / 5.0);
			t1 = (-1.0 / 12.0) * w * (t + 4.0);
			xw[2] = t0 + t1;
			xw[3] = t0 - t1;
			t0 = (1.0 / 16.0) * (9.0 / 5.0 - t);
			t1 = (1.0 / 24.0) * w * (w4 - w2 - 5.0);
			xw[1] = t0 + t1;
			xw[4] = t0 - t1;
			/* y */
			w = y - (double)yi[2];
			w2 = w * w;
			yw[5] = (1.0 / 120.0) * w * w2 * w2;
			w2 -= w;
			w4 = w2 * w2;
			w -= 1.0 / 2.0;
			t = w2 * (w2 - 3.0);
			yw[0] = (1.0 / 24.0) * (1.0 / 5.0 + w2 + w4) - yw[5];
			t0 = (1.0 / 24.0) * (w2 * (w2 - 5.0) + 46.0 / 5.0);
			t1 = (-1.0 / 12.0) * w * (t + 4.0);
			yw[2] = t0 + t1;
			yw[3] = t0 - t1;
			t0 = (1.0 / 16.0) * (9.0 / 5.0 - t);
			t1 = (1.0 / 24.0) * w * (w4 - w2 - 5.0);
			yw[1] = t0 + t1;
			yw[4] = t0 - t1;
			break;
		default:
			dp (0,"Invalid spline degree\n");
	}

	/* apply the mirror boundary conditions */

	for (k = 0; k <= SplDeg; k++)
	{
		xi[k] = (width == 1) ? (0) : ((xi[k] < 0) ?
			(-xi[k] - Width2 * ((-xi[k]) / Width2))
			: (xi[k] - Width2 * (xi[k] / Width2)));

		if (width <= xi[k])
		{
			xi[k] = Width2 - xi[k];
		}

		yi[k] = (height == 1) ? (0) : ((yi[k] < 0) ?
			(-yi[k] - Height2 * ((-yi[k]) / Height2))
			: (yi[k] - Height2 * (yi[k] / Height2)));

		if (height <= yi[k])
		{
			yi[k] = Height2 - yi[k];
		}
	}


}


/* Initialize the causal coefficients. Pass the sample values in c
   and the number of samples in ns. The actual pole is in z and the
   admissible relative error in epsilon.
*/


double	init_causal (double* c, int ns, double z, double epsilon)
{
double	Sum, zn, z2n, iz;
long	n, Horizon;

	/* this initialization corresponds to mirror boundaries */
	Horizon = ns;
	if (epsilon > 0.0)
	{
		Horizon = (long)ceil(log(epsilon) / log(fabs(z)));
	}

	if (Horizon < ns) 	/* accelerated loop */
	{
		zn = z;
		Sum = c[0];
		for (n = 1; n < Horizon; n++)
		{
			Sum += zn * c[n];
			zn *= z;
		}
		return Sum;
	}
	else	/* full loop */
	{
		zn = z;
		iz = 1.0 / z;
		z2n = pow(z, (double)(ns - 1));
		Sum = c[0] + z2n * c[ns - 1];
		z2n *= z2n * iz;
		for (n = 1L; n <= ns - 2; n++)
		{
			Sum += (zn + z2n) * c[n];
			zn *= z;
			z2n *= iz;
		}
		return Sum / (1.0 - zn * zn);
	}
}


/* Get initial anticausal coefficient. Parameters as above */

double	init_anticausal (double* c, int ns, double z)
{
	return ((z / (z * z - 1.0)) * (z * c[ns - 2] + c[ns - 1]));
}



/* Compute the interpolation coefficients. Input the sample values
   in c[]. C[] contains the coefficients on exit. Specify the number of
   samples (and coefficients) in ns, the poles in z[] and the
   number of poles in np and the
   admissible relative error in epsilon.
*/


void get_interpol_coeff (double* c, int ns, double* z, int np, double epsilon)
{

double	Lambda = 1.0;
int	n, k;


	if (ns == 1) return; /* special case required by mirror boundaries */

	for (k = 0; k < np; k++) 	/* compute the overall gain */
	{
		Lambda = Lambda * (1.0 - z[k]) * (1.0 - 1.0 / z[k]);
	}

	for (n = 0; n < ns; n++) /* apply the gain */
	{
		c[n] *= Lambda;
	}

	for (k = 0L; k < np; k++) /* loop over all poles */
	{
		c[0] = init_causal (c, ns, z[k], epsilon);	/* causal initialization */

		for (n = 1; n < ns; n++) 	/* causal recursion */
		{
			c[n] += z[k] * c[n - 1];
		}

		c[ns - 1] = init_anticausal (c, ns, z[k]); 	/* anticausal initialization */

		for (n = ns - 2; n >= 0; n--) /* anticausal recursion */
		{
			c[n] = z[k] * (c[n + 1] - c[n]);
		}
	}
}


/* Compute the coefficients for one image plane (2D only). This is
   and in-place operation on img, where img must be pre-allocated
   and be of type B_FLOAT. The original image values must be copied
   into img->data. SplDeg is the spline degree.
*/


void compute_coeff (image_type* img, int SplDeg)
{
double	*Line;
float* p;
double	Pole[2];
int np;
int x,y,idx;

	g_assert (img->imgtype == B_FLOAT);
	p = img->data;

	/* recover the poles from a lookup table. If an invalid spline degree
	   is given, pretend we wanted degree 2 */
	switch (SplDeg)
	{
		case 2:
			np = 1; Pole[0] = sqrt(8.0) - 3.0;
			break;
		case 3:
			np = 1; Pole[0] = sqrt(3.0) - 2.0;
			break;
		case 4:
			np = 2;
			Pole[0] = sqrt(664.0 - sqrt(438976.0)) + sqrt(304.0) - 19.0;
			Pole[1] = sqrt(664.0 + sqrt(438976.0)) - sqrt(304.0) - 19.0;
			break;
		case 5:
			np = 2;
			Pole[0] = sqrt(135.0 / 2.0 - sqrt(17745.0 / 4.0)) + sqrt(105.0 / 4.0)
				- 13.0 / 2.0;
			Pole[1] = sqrt(135.0 / 2.0 + sqrt(17745.0 / 4.0)) - sqrt(105.0 / 4.0)
				- 13.0 / 2.0;
			break;
		default:
			np = 1; Pole[0] = sqrt(8.0) - 3.0;
			dp (0,  "Invalid spline degree selected\n");
	}

	/* Sample computation is an in-place separable process, so perform
	the operation twice; once along X then along Y. The tolerance epsilon
	is defined in math.h and is somewhere around 2.22e-16 */

	Line = calloc(img->xmax, sizeof(double));
	if (!Line) fatal_memerr();

	for (y = 0; y < img->ymax; y++)
	{
		idx = img->xmax * y;
		for (x=0; x<img->xmax; x++) Line[x] = p[x+idx];
		get_interpol_coeff (Line, img->xmax, Pole, np, DBL_EPSILON);
		for (x=0; x<img->xmax; x++) p[x+idx] = Line[x];
	}
	free(Line);

	Line = calloc(img->ymax, sizeof(double));
	if (!Line) fatal_memerr();

	for (x = 0; x < img->xmax; x++)
	{
		idx=0;
		for (y=0; y<img->ymax; y++) { Line[y] = p[x+idx]; idx+=img->xmax; }
		get_interpol_coeff (Line, img->ymax, Pole, np, DBL_EPSILON);
		idx=0;
		for (y=0; y<img->ymax; y++) { p[x+idx] = Line[y]; idx+=img->xmax; }
	}
	free(Line);
}


/* Here is the actual 2D interpolation function. It requires the coefficients
   to be pre-computed and cached in the "image" coeff. It does *NOT* require
   the original image data as this was already used to compute the coefficients.*/


double spline_interpol_core (image_type* coeff, float x, float y, int SplDeg)
{
float	*p;
double	xWeight[6], yWeight[6];
double	interpolated;
double	w;
int		xIndex[6], yIndex[6];
int	i, j, idx;


	g_assert (coeff->imgtype == B_FLOAT);
	p = coeff->data;

	/* Indices and weights need to be computed for every x,y coordinate */

	compute_weights (x,y, xIndex, yIndex, xWeight, yWeight, coeff->xmax, coeff->ymax, SplDeg);

	/* perform interpolation */

	interpolated = 0.0;
	for (j = 0; j <= SplDeg; j++)
	{
		idx = yIndex[j]* coeff->xmax;
		w = 0.0;
		for (i = 0; i <= SplDeg; i++)
		{
			w += xWeight[i] * p[xIndex[i] + idx];
		}
		interpolated += yWeight[j] * w;
	}

	return(interpolated);
}


/* The spline interpolation routine with coefficient caching mechanism */
/* TODO: Can't handle RGB */

double spline_interpol (image_type* img, float xf, float yf, int z, int SplDeg)
{
int x,y;
float* p;

	if (!spl_cache_invalidate)		/* Check if we need to invalidate */
	{
		if (spl_cache_z != z) spl_cache_invalidate=1;
		if (spl_cache_deg != SplDeg) spl_cache_invalidate=1;
		if (spl_cache_uid != img->uid) spl_cache_invalidate=1;

		if (spl_cache_invalidate)
		{
		 	freebuf (&splcoeff);
			dp (1, "Spline interpolation: Cache invalidated\n");
		}
	}

	if (spl_cache_invalidate)		/* Need to recompute the coefficients */
	{

		dp (1, "Spline interpolation: Recomputing coefficients\n");
		if (allocate_image (&splcoeff, B_FLOAT, img->xmax, img->ymax, 1)==-1)
			fatal_memerr();
		p = splcoeff.data;

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				p[x + img->xmax*y] = readbuf_flt (img, x,y,z);

		compute_coeff (&splcoeff, SplDeg);

		spl_cache_invalidate=0;
		spl_cache_z = z;
		spl_cache_deg=SplDeg;
		spl_cache_uid = img->uid;
	}

	return spline_interpol_core (&splcoeff, xf,yf, SplDeg);

}




/******************************************************************/


