
/*
 *	Spider Op: My old vertebral cortex approximation / segmentation algorithm
	See Haidekker MA et al, Brit J Radiol 1997; 70: 586-93.
 */



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include "spiderop.h"
#include "roi.h"
#include "dlgbox.h"
#include "imgbuf.h"
#include "application.h"
#include "secondwnd.h"
#include "filter.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "functions.h"
#include "vertanalysis.h"


#define MAXVERTICES 	360
#define INITIAL_ODIAM 	150
#define MAXGAUSS		128


#define EDET_GRAD		0
#define EDET_MAX		1
#define	EDET_MIN		2
#define	EDET_REL		3
#define EDET_LOCVAR		4

vertex spid_v[MAXVERTICES];	/* the vertices of the polygon found */
float gaussfn[MAXGAUSS];	/* No more than 128 points, irrespective of actual width */

GtkAdjustment* spider_idiam_adj;
GtkAdjustment* spider_odiam_adj;
GtkAdjustment *spider_vertex_spin_adj;
GtkAdjustment* spider_reject_adj;
GtkAdjustment* spider_rel_thresh_adj;
GtkAdjustment* spider_prefilter_adj;
GtkAdjustment *spider_followmax_adj;
GtkAdjustment *spider_boxsize_adj;
GtkAdjustment *spider_decay_adj;

GtkWidget *spider_statusbar;
GtkWidget *spider_label10;
char spider_statusbar_was_used;
int edet_function;	/* Edge detector, "Gradient", "Max", "Min", "Rel" */
int lvl2filter;		/* Post-processing filter in level 2, "none", "weak", "medium", "agressive" */

/*	Ray Stuctures and other globals */
/* NOW IN DEFS.H */



float reject_thresh;
float rel_thresh;
float prefilter_depth;
int initial_diam;
int idiam;
int odiam;
int old_odiam;
int spider_dirflg;
int spider_lastlevel;

int analysis;		/* Which analysis to use: Vertebrae or vessels */






/**************************************************************

	Spider Op Algorithms:
	spider_create_image creates the image section to work on
	by cloning one slice from the main image.
	Note - as of 0.9.0, we clone the entire slice.
	spider_pass1 does the actual edge detection and creates
				a preliminary array of vertices
	spider_pass2 post-processes the vertices (filters)
	spider_draw_results gives some optical feedback by drawing
	           the sample traces and the polygon into the 
			   secondary display
	
***************************************************************/


vertex spider_create_image (image_type* img, int radius, vertex center)
{
int bx1,by1,bx2,by2;		/* bounding box */
int x,y, xf,yf;
float* p1;
float f,fmax,fmin;
vertex v;
long idx;
		
	/* Crop and clone the main image; make FLOAT */

	bx1 = center.x-radius; bx2 = center.x+radius;
	by1 = center.y-radius; by2 = center.y+radius;
	
	/* Observe original image limits. If we do an adjustment,
	   we also need to adjust the offset "centerglobal".
		The following code is obsolete and should be deleted on occasion. */

	if ((bx1<0) && (bx2>=mainimg.xmax))	/* doesn't fit at all */
	{
		bx1 = 0; bx2 = mainimg.xmax; 
		v.x=0; centerglobal.x = center.x;
	}
	else if (bx1<0)						/* only left-hand side adjusted */
	{
		bx1 = 0;
		v.x=0; centerglobal.x = bx2-radius;
	}
	else if (bx2>=mainimg.xmax)		/* only right-hand side adjusted */
	{
		bx2 = mainimg.xmax-1;
		v.x=bx1; centerglobal.x = radius;
	}
	else							/* no adjustment - circle fits within mainimg */
	{
		v.x=bx1; centerglobal.x = (bx2-bx1)/2;
	}

	/* ... and the same for Y */

	if ((by1<0) && (by2>=mainimg.ymax))	/* doesn't fit at all */
	{
		by1 = 0; by2 = mainimg.ymax; 
		v.y=0; centerglobal.y = center.y;
	}
	else if (by1<0)						/* only left-hand side adjusted */
	{
		by1 = 0;
		v.y=0; centerglobal.y = by2-radius;
	}
	else if (by2>=mainimg.ymax)		/* only right-hand side adjusted */
	{
		by2 = mainimg.ymax-1;
		v.y=by1; centerglobal.y = radius;
	}
	else							/* no adjustment - circle fits within mainimg */
	{
		v.y=by1; centerglobal.y = (by2-by1)/2;
	}

	/* Instead, enable this as of 0.9.0 */

	bx1=0; by1=0; bx2=511; by2=511;
	centerglobal.x=center.x; centerglobal.y=center.y; 
	v.x=0; v.y=0;

	xf = 1+bx2-bx1; yf=1+by2-by1;		/* size */
	if (allocate_image (img, B_FLOAT, mainimg.xmax, mainimg.ymax, 1)==-1)
		fatal_memerr ();

	copybuf (&mainimg, img, 0,0,cur_slice, 0,0, 0, mainimg.xmax*mainimg.ymax);
	make_minmax (img);

	dp (1, "Spiderop: Secondary image created. Center is %d,%d radius %d\n",
					center.x,center.y,radius);
	dp (1, "          Bounding box (%d %d) to (%d %d)  -> offset (%d %d), center (%d %d)\n",
					bx1,by1,bx2,by2,v.x,v.y, centerglobal.x, centerglobal.y);

	return v;

}


/*******************************************************

	Functions used by Pass 1

*******************************************************/

/* Convolve the gray values with a Gaussian function of width a.
   This acts as a nice variable-strength low-pass filter.
   We prepare the Gauss function array beforehand to save time
   - otherwise we'd have to compute the exp() term for each ray
*/


int gauss_prep (float a)
{
int i,j;
float twoasquare, prefax;

	if (a==0.0)
	{
		gaussfn[0]=1;
		return 1;
	}

	twoasquare = 1/(2*a*a);	/* dividing takes longer than multiplying -- division before loop */
	prefax = 1/(a*sqrt(2*PI));
	j=0;
	
	for (i=0; i<MAXGAUSS; i++)
	{
		gaussfn[i] = exp(-i*i*twoasquare)*prefax;
		if (gaussfn[i] > 0.01) 						/* Cutoff value */
			j=i;
		else
			gaussfn[i]=0;
	}

	return j;

}


void gauss_convolve (ray* aray, int convwidth, char do_graph)
{
int i,j,k;
float convolved[aray->numPixels];
float c;
	
	
	for (j = 0; j < aray->numPixels; j++)
	{
		c = 0;
		for (i=0; i<convwidth; i++)
		{
			k = j-i; if (k<0) k=0;
			c += gaussfn[i] * aray->pixelValues[k];
			k = j+i; if (k>=aray->numPixels) k = aray->numPixels-1;
			c += gaussfn[i] * aray->pixelValues[k];
		}
		convolved[j] = c;
	}			

	for (j = 0; j < aray->numPixels; j++)
		aray->pixelValues[j] = convolved[j];

}


float locvar (image_type* img, int x, int y, int z, int neigh)
{
int i,j,n;
float r, Sx, Sxx;

	n=0; Sx=0; Sxx=0;
	for (i=y-neigh; i<=y+neigh; i++)
	{
		if ((y>=0) && (y<img->ymax))
			for (j=x-neigh; j<=x+neigh; j++)
			{
				if ((x>=0) && (x<img->xmax))
				{
					r = readbuf_flt (img, i,j,z);
					Sx += r; Sxx+=SQR(r); n++;
				}
			}
	}
	
	if (n==0) return 0;
	return sqrt( Sxx/n - SQR(Sx/n) );
}



/* BresLine computes the ray trace (array of x/y coords) and samples the gray values */

int BresLine(ray *aray)
{
int Ax = aray->startx;
int Ay = aray->starty;
int Bx = aray->endx;
int By = aray->endy;
 
int counter= 0;
 

	/* INITIALIZE THE COMPONENTS OF THE ALGORITHM THAT ARE NOT AFFECTED BY THE
    	SLOPE OR DIRECTION OF THE LINE */
	int dX = abs(Bx-Ax);    /* store the change in X and Y of the line endpoints */
	int dY = abs(By-Ay);
        
	int CurrentX = Ax;              // store the starting point (just point A)
	int CurrentY = Ay;
        
	/* DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION) */
	int Xincr, Yincr;
	if (Ax > Bx) { Xincr=-1; } else { Xincr=1; }    /* which direction in X? */
	if (Ay > By) { Yincr=-1; } else { Yincr=1; }    /* which direction in Y? */
        
	/* DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
		AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
		ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE. */
    
	if (dX >= dY)   /* if X is the independent variable */
	{           
		int dPr         = dY<<1;              /* amount to increment decision if right is chosen (always) */
		int dPru        = dPr - (dX<<1);      /* amount to increment decision if up is chosen */
		int P           = dPr - dX;           /* decision variable start value */

		for (; dX>=0; dX--)                   /* process each point in the line one at a time (just use dX) */
		{
			aray->XPixelCoords[counter] = CurrentX;
			aray->YPixelCoords[counter] = CurrentY;
			if (edet_function==EDET_LOCVAR)
				aray->pixelValues[counter] = locvar (&spider_img, CurrentX + centerglobal.x, CurrentY+centerglobal.y, 0,3);
			else
				aray->pixelValues[counter] = readbuf_flt (&spider_img, CurrentX + centerglobal.x, CurrentY+centerglobal.y, 0);
			/* aray->pixelValues[counter] = s_im->rgb_data[3*(CurrentX + centerglobal.x + secw*(CurrentY+centerglobal.y))]; */
			counter++;

			if (P > 0)                           /* is the pixel going right AND up? */
			{ 
				CurrentX+=Xincr;             /* increment independent variable */
				CurrentY+=Yincr;             /* increment dependent variable */
				P+=dPru;                     /* increment decision (for up) */
			}
			else                                 /* is the pixel just going right? */
			{
				CurrentX+=Xincr;             /* increment independent variable */
				P+=dPr;                      /* increment decision (for right) */
			}
		}
	}
	else           /* if Y is the independent variable */
	{
		int dPr         = dX<<1;                     /* amount to increment decision if right is chosen (always) */
		int dPru        = dPr - (dY<<1);             /* amount to increment decision if up is chosen */
		int P           = dPr - dY;                  /* decision variable start value */

		for (; dY>=0; dY--)                          /* process each point in the line one at a time (just use dY) */
		{
			aray->XPixelCoords[counter] = CurrentX;
			aray->YPixelCoords[counter] = CurrentY;
			if (edet_function==EDET_LOCVAR)
				aray->pixelValues[counter] = locvar (&spider_img, CurrentX + centerglobal.x, CurrentY+centerglobal.y, 0,3);
			else
				aray->pixelValues[counter] = readbuf_flt (&spider_img, CurrentX + centerglobal.x, CurrentY+centerglobal.y, 0);
			/* aray->pixelValues[counter] = s_im->rgb_data[3*(CurrentX + centerglobal.x + secw*(CurrentY+centerglobal.y))]; */
			counter++;

			if (P > 0)                           /* is the pixel going up AND right? */
			{ 
				CurrentX+=Xincr;             /* increment dependent variable */
				CurrentY+=Yincr;             /* increment independent variable */
				P+=dPru;                     /* increment decision (for up) */
			}
			else                                 /* is the pixel just going up? */
			{
				CurrentY+=Yincr;             /* increment independent variable */
				P+=dPr;                      /* increment decision (for right) */
			}
		}               
	}               

	return counter;
}




void  ComputeRaystoConnect (ray rayforv[], int numv)  
{
int i;	
		/* 	Need  connect array last point to first point (wrap around) outside of loop */
	
	raystoconnectvtx[numvertices-1].startx = rayforvertices[numvertices-1].vtxXPixelCoord;
	raystoconnectvtx[numvertices-1].starty = rayforvertices[numvertices-1].vtxYPixelCoord;
	raystoconnectvtx[numvertices-1].endx = rayforvertices[0].vtxXPixelCoord;
	raystoconnectvtx[numvertices-1].endy = rayforvertices[0].vtxYPixelCoord;
	raystoconnectvtx[numvertices-1].numPixels = BresLine(&raystoconnectvtx[numvertices-1]);
	
	for (i = 0; i< numvertices-1; i++)
	{
		raystoconnectvtx[i].startx = rayforvertices[i].vtxXPixelCoord;
		raystoconnectvtx[i].starty = rayforvertices[i].vtxYPixelCoord;
		raystoconnectvtx[i].endx = rayforvertices[i+1].vtxXPixelCoord;
		raystoconnectvtx[i].endy = rayforvertices[i+1].vtxYPixelCoord;			
		raystoconnectvtx[i].numPixels = BresLine(&raystoconnectvtx[i]);
	}  

}
				


int maximum (float *set, int size, int flag)	/* returns INDEX of max array element of array "set" */
{
int maxPt = 0; /*Index of Maximum array element*/
int j = 0;
	
	if (flag == 0) 
	{
		for (j = 1; j <size; j ++)
			if (set[j] > set[maxPt])
				maxPt = j;
	}

	
	if (flag == 1) 
	{
		maxPt = size-1;
		for (j = size-2; j > 0; j --)
			if (set[j] > set[maxPt])
				maxPt = j;
	}	
	
	return maxPt;
}
			

int minimum (float *set, int size, int flag)	/* returns INDEX of max array element */
{
int minPt = 0; /*Index of Minimum array element*/
int j = 0;
	
	if (flag == 0) 
	{
		for (j = 1; j <size; j ++)
			if (set[j] < set[minPt])
				minPt = j;
	}
		
	if (flag == 1) 
	{
		minPt = size-1;
		for (j = size-2; j > 0; j --)
			if (set[j] < set[minPt])
				minPt = j;
	}	
		
		
	return minPt;
}
	



void gradient_detect (ray *rayforspi, int flag) 
{
int j,n;
int tempindex;
float gradArray[rayforspi->numPixels];
	
	n = rayforspi->numPixels;
	for (j=0; j<n-1; j++) gradArray[j]=0;

	/* First take forward differences on left and right edges */

	gradArray[0] = fabs(rayforspi->pixelValues[1] - rayforspi->pixelValues[0] ) ;
	gradArray[n-1] = fabs(rayforspi->pixelValues[n-1] - rayforspi->pixelValues[n-2]);

	/* central differences for the rest */
	for (j = 0; j < n-2; j++)
		gradArray[j+1] = fabs ((rayforspi->pixelValues[j+2] - rayforspi->pixelValues[j])/2.0 ) ;

	tempindex = maximum(gradArray, n, flag);

	rayforspi->vtxPixelIndex = tempindex;
	rayforspi->vtxXPixelCoord = rayforspi->XPixelCoords[tempindex];
	rayforspi->vtxYPixelCoord = rayforspi->YPixelCoords[tempindex];
	rayforspi->vtxPixelValue = readbuf_flt (&spider_img, rayforspi->vtxXPixelCoord+centerglobal.x,
														rayforspi->vtxYPixelCoord+centerglobal.y, 0);

}


void maximum_detect(ray *rayforspi, int flag) 
{
int tempindex;
	
	rayforspi->vtxPixelIndex = maximum(rayforspi->pixelValues, rayforspi->numPixels, flag);
	tempindex = rayforspi->vtxPixelIndex;	
	rayforspi->vtxXPixelCoord = rayforspi->XPixelCoords[tempindex];
	rayforspi->vtxYPixelCoord = rayforspi->YPixelCoords[tempindex];
	rayforspi->vtxPixelValue = readbuf_flt (&spider_img, rayforspi->vtxXPixelCoord+centerglobal.x,
														rayforspi->vtxYPixelCoord+centerglobal.y, 0);
	/* rayforspi->vtxPixelValue = s_im->rgb_data[  3*(rayforspi->vtxXPixelCoord+centerglobal.x + 
								(rayforspi->vtxYPixelCoord+centerglobal.y)*secw)]; */

}			


void minimum_detect(ray *rayforspi, int flag) 
{
int tempindex;

	rayforspi->vtxPixelIndex = minimum(rayforspi->pixelValues, rayforspi->numPixels, flag);
	tempindex = rayforspi->vtxPixelIndex;	
	rayforspi->vtxXPixelCoord = rayforspi->XPixelCoords[tempindex];
	rayforspi->vtxYPixelCoord = rayforspi->YPixelCoords[tempindex];
	rayforspi->vtxPixelValue = readbuf_flt (&spider_img, rayforspi->vtxXPixelCoord+centerglobal.x,
														rayforspi->vtxYPixelCoord+centerglobal.y, 0);
	/* rayforspi->vtxPixelValue = s_im->rgb_data[  3*(rayforspi->vtxXPixelCoord+centerglobal.x + 
								(rayforspi->vtxYPixelCoord+centerglobal.y)*secw)]; */

}							
							
	
void relative_detect (ray *rayforspi, int flag)	
{
int i,j;
int tempindex, tempflag = 0;
float constant, runSum, curAvg, upbnd, lowbnd,v;	
	
	constant = 1.0 + (rel_thresh/100.0);
	tempflag = 0;
	runSum = 0;						

	for (i = 0; i < rayforspi->numPixels; i++) 
	{
		j=i; if (flag) j=rayforspi->numPixels - 1 - i;  /* in -> out versus out -> in */
		v = fabs(rayforspi->pixelValues[j]);
		runSum += v;
		curAvg = runSum/(i+1); 
		upbnd = fabs(constant*curAvg); lowbnd = fabs(curAvg/constant);
		
		if ((tempflag==0) && ((v > upbnd) || (v < lowbnd)))
		{
			rayforspi->vtxPixelIndex = j;
			tempindex = rayforspi->vtxPixelIndex;	
			rayforspi->vtxXPixelCoord = rayforspi->XPixelCoords[tempindex];
    		rayforspi->vtxYPixelCoord = rayforspi->YPixelCoords[tempindex];
			rayforspi->vtxPixelValue = readbuf_flt (&spider_img, rayforspi->vtxXPixelCoord+centerglobal.x,
													rayforspi->vtxYPixelCoord+centerglobal.y, 0);
			/* rayforspi->vtxPixelValue = s_im->rgb_data[  3*(rayforspi->vtxXPixelCoord+centerglobal.x + 
							(rayforspi->vtxYPixelCoord+centerglobal.y)*secw)]; */
			tempflag = 1;
		}
	}		


	/* IF after either of these (flag == 0 and flag == 1) no pixels have been set, we need to set them to something
		otherwise the old values will remain (from whatever other edge detection technique) and we do not
		 want that.  */

	if (tempflag == 0)
	{
		rayforspi->vtxPixelIndex = 0; if (flag) rayforspi->vtxPixelIndex = rayforspi->numPixels - 1;
		tempindex = rayforspi->vtxPixelIndex;	
		rayforspi->vtxXPixelCoord = rayforspi->XPixelCoords[tempindex];
		rayforspi->vtxYPixelCoord = rayforspi->YPixelCoords[tempindex];
		rayforspi->vtxPixelValue = readbuf_flt (&spider_img, rayforspi->vtxXPixelCoord+centerglobal.x,
													rayforspi->vtxYPixelCoord+centerglobal.y, 0);
		/* rayforspi->vtxPixelValue = s_im->rgb_data[  3*(rayforspi->vtxXPixelCoord+centerglobal.x + 
							(rayforspi->vtxYPixelCoord+centerglobal.y)*secw)]; */
	}

}											


void compute_outside_average (ray* aray)
{
int i, cnt;
double avg;

	avg=0; cnt=0;
	for (i=aray->vtxPixelIndex+1; i<aray->numPixels; i++)
	{
		avg+=aray->pixelValues[i]; cnt++;
	}
	if (cnt==0) 
		aray->vtxOutsideValue = 0;
	else
		aray->vtxOutsideValue = avg/cnt;

}


void FindVertexPoints (ray *aray, int flag)
{

	/* The flag lets you state which direction you want to the search to start from - 
		from inner diameter to outer diameter or vice versa.  */
	/* Right now just deal with the MAXIMUM as the threshold detection level */
	switch (edet_function)   /* 1 Indicates maximum */
	{
		case EDET_GRAD:		/*   Gradient */
			gradient_detect (aray, flag);
			break;
				
		case EDET_MAX:		/* Maximum		*/	
			maximum_detect (aray, flag);
			break;
					
		case EDET_MIN:		/* Minimum			 */
			minimum_detect (aray, flag);
			break;
					
		case EDET_REL:		/* Relative: compute running average , and choose a point that above that avg*const. */
		case EDET_LOCVAR:	/* Local variance, as relative, but using loc var instead of pixel value */

			relative_detect(aray, flag);
			break;
					
		default:
			break;
											 
	}
			
	compute_outside_average (aray);
	aray->vtxRadialPosition = sqrt(SQR(aray->vtxXPixelCoord) + 
			SQR(aray->vtxYPixelCoord)  );
			
}

	


/*********************************************************

	Functions used by pass 2

*********************************************************/


void filter_data() 
{
int i,j,n,n2;
float tempforfilter[MAXVERTICES];
/* FILE *fid;		*/
char weights[5];
float a,b;
	
	if (lvl2filter==0) return;
	switch ( lvl2filter ) 
	{ 
		case 1:
			weights[0]=1; weights[1]=2; weights[2]=1; weights[3]=0; weights[4]=0;
			n=3; n2=1;
			break;
		case 2:
			weights[0]=1; weights[1]=4; weights[2]=6; weights[3]=4; weights[4]=1;
			n=5; n2=2;
			break;
		case 3:
			weights[0]=1; weights[1]=1; weights[2]=1; weights[3]=1; weights[4]=1;
			n=5; n2=2;
			break;
		default:
			weights[0]=1; weights[1]=2; weights[2]=1; weights[3]=0; weights[4]=0;
			n=3; n2=1;
			break;
	}
	
		
	/* fid = fopen("test.txt", "w"); */

	for (i=0; i<numvertices; i++)
	{
    	a=0; b=0;
        for (j=0; j<n; j++)
        {
			/* Always be aware that the -1 (below) is dependent on size of filter.  If there
					are three weights it must be 1, if five, 2, if seven 3. This is considered in the presetting 
					of the variables n (filter size) and n2 (filter shoulder width) */

			a += rayforvertices[ (i+numvertices+j-n2) % numvertices ].vtxRadialPosition * weights[j];
			b += weights[j];
		}
		tempforfilter[i] = a/b;
		/* fprintf (fid, "%f\t%f\n", rayforvertices[i].vtxRadialPosition, tempforfilter[i]); */
	}

	/* fclose(fid); */

	for (i=0; i<numvertices; i++)
	{
		rayforvertices[i].vtxRadialPosition = tempforfilter[i];
	}
	
	
}

	

void use_rejection_threshold ()
{
			
int i, j=0;
float curAvg;
int counter = 0 ;  /* Have this start at 4, since we are not messing with the first
							two points and last two points. */
float leeway;
			
	/* 	Lets not mess with the two first points and two last points in the loop, since they have do not 
		have all four nearest neighbors.  */

	if (reject_thresh > 0.0) 
	{
												
		for (i = 0; i <  numvertices; i ++) 
		{

			/* curAvg is the average of the current point's 4 neighbors */
			curAvg = (float)(rayforspider[(i+numvertices-2) % numvertices].vtxRadialPosition +
							rayforspider[(i+numvertices-1) % numvertices].vtxRadialPosition +
							rayforspider[(i+numvertices+1) % numvertices].vtxRadialPosition +
							rayforspider[(i+numvertices+2) % numvertices].vtxRadialPosition)/4;
					
			leeway = (1-reject_thresh)*curAvg;
			if ( fabs( rayforspider[i].vtxRadialPosition - curAvg) > leeway) 
			{
				// Its outside of range and should be flagged. 
				rayforspider[i].validVertex = 1;
			}
			else
			{
				rayforspider[i].validVertex = 0;
				counter++;
			}
		}

	}
	else	/* No rejection threshold used, accept all vertices */
	{
			
		for (i = 0; i < numvertices; i++)
			rayforspider[i].validVertex = 0;
			
		counter = numvertices;

	}

				
	for (i = 0; i < numvertices; i++)
	{
		if (rayforspider[i].validVertex == 0)
		{
			rayforvertices[j] = rayforspider[i];
			j++;
		}
	}

	/* 	So now rayforvertices should just have the vertices we want should have that number of elements.
		i.e should be the size of counter.  */

	numvertices = counter;		/* New number of vertices (the ones that will actually be used)
											always <= numspiderlegs */

}		



void convert_radialtocartesian (ray *rayforv, int numv)	
{
int k;
	
	/* This is the simple convertion using x = r*cos phi and y = r * sin phi */

	for (k = 0; k < numv; k++) 
	{
		rayforv[k].vtxXPixelCoord = IROUND(rayforv[k].vtxRadialPosition*rayforv[k].rotmatrix[0][0]);
		rayforv[k].vtxYPixelCoord = IROUND(rayforv[k].vtxRadialPosition*rayforv[k].rotmatrix[0][1]);
	}
}




/*******************************************************************************

	To get an optical feedback, draw the spider rays & polygon
	into the secondary display image

*******************************************************************************/

#define VXI(I)	rayforspid[(I)].vtxXPixelCoord
#define VYI(I)	rayforspid[(I)].vtxYPixelCoord

vertex center_of_gravity (ray *rayforspid, int numv)
{
float cogx, cogy, area, r;
vertex cog;
int i,j;

	/* Old and wrong algorithm, keep it - may be more benign 
	cogx=cogy=0;
	for (i=0; i<numv; i++)
	{
		cogx += rayforspid[i].vtxXPixelCoord;
		cogy += rayforspid[i].vtxYPixelCoord;
	}
	cog.x = (int) (0.5+cogx/numv);
	cog.y = (int) (0.5+cogy/numv);
	*/
	
	area = 0; cogx=0; cogy=0;
	for (i=0; i<numv; i++)
	{
		j=i+1; if (j>=numv) j=0;	/* wrap-around */
		r = VXI(i)*VYI(j) - VXI(j)*VYI(i);
		area += r;
		cogx += (VXI(i)+VXI(j))*r;
		cogy += (VYI(i)+VYI(j))*r;
	}
	/* Post-adjust */
	area = 0.5*area;
	cogx /= 6*area;
	cogy /= 6*area;
	
	cog.x = (int) (0.5+cogx);
	cog.y = (int) (0.5+cogy);


	return cog;
}


#undef VXI
#undef VYI


void HighlightVertices (ray *rayforspider, int numv, rgbtriplet color)
{

int i,x,y, xmax,ymax,stride;
int idx;
unsigned char *s_im_buf;

	xmax = gdk_pixbuf_get_width (s_img.pixbuf);
	ymax = gdk_pixbuf_get_height (s_img.pixbuf);
	stride = gdk_pixbuf_get_rowstride (s_img.pixbuf);
	s_im_buf = gdk_pixbuf_get_pixels (s_img.pixbuf);			/* Pointer to pixbuf's rgb data */

	/* Now make the actual point picked by whatever method of edge detection a different color */
	for (i = 0; i < numv; i++)
	{
		x = rayforspider[i].vtxXPixelCoord + centerglobal.x;
		y = rayforspider[i].vtxYPixelCoord + centerglobal.y;
		if ((x>=0) && (x<s_img.xmax) && (y>=0) && (y<s_img.ymax))
		{
			idx= 3*x + stride*y;
			s_im_buf[idx++] = color.red;
			s_im_buf[idx++] = color.green;
			s_im_buf[idx++] = color.blue;
		}
	}

	return;

}



void DrawSpider(ray rayforspider[], int numv, rgbtriplet color)
{
int i,j,x,y, xmax, ymax, stride;
int idx;
unsigned char *s_im_buf;


	xmax = gdk_pixbuf_get_width (s_img.pixbuf);
	ymax = gdk_pixbuf_get_height (s_img.pixbuf);
	stride = gdk_pixbuf_get_rowstride (s_img.pixbuf);
	s_im_buf = gdk_pixbuf_get_pixels (s_img.pixbuf);			/* Pointer to pixbuf's rgb data */


	/* First draw the rays */
	for (i = 0; i < numv; i++)
	{
		for (j = 0; j<rayforspider[i].numPixels; j++)
		{
			x = rayforspider[i].XPixelCoords[j] + centerglobal.x;
			y = rayforspider[i].YPixelCoords[j] + centerglobal.y;
			if ((x>=0) && (x<s_img.xmax) && (y>=0) && (y<s_img.ymax))
			{
				idx= 3*x + stride*y;
				s_im_buf[idx++] = color.red;
				s_im_buf[idx++] = color.green;
				s_im_buf[idx++] = color.blue;
			}
		}

	}

	return;
}


void mark_cog (ray* rayforspid, int numv, rgbtriplet color)
{
int idx, xmax, ymax, stride,x,y;
vertex cog;
unsigned char *s_im_buf;

	xmax = gdk_pixbuf_get_width (s_img.pixbuf);
	ymax = gdk_pixbuf_get_height (s_img.pixbuf);
	stride = gdk_pixbuf_get_rowstride (s_img.pixbuf);
	s_im_buf = gdk_pixbuf_get_pixels (s_img.pixbuf);			/* Pointer to pixbuf's rgb data */

		
	cog = center_of_gravity (rayforspid, numv);
	cog.x += centerglobal.x;
	cog.y += centerglobal.y;

	for (y=cog.y-1; y<=cog.y+1; y++)
	{
		idx = 3*cog.x + stride*y;
		s_im_buf[idx++] = color.red;
		s_im_buf[idx++] = color.green;
		s_im_buf[idx] = color.blue;
	}

	for (x=cog.x-1; x<=cog.x+1; x++)
	{
		idx = 3*x + stride*cog.y;
		s_im_buf[idx++] = color.red;
		s_im_buf[idx++] = color.green;
		s_im_buf[idx] = color.blue;
	}

}


void spider_draw_results() 
{
rgbtriplet red = {0xff, 0x00, 0x00};
rgbtriplet yellow = {0xff, 0xff, 0x00};
rgbtriplet ltblue = {0x00, 0x7f, 0xff};
		
		DrawSpider(rayforspider, numspiderlegs, red);  		/* Draws the actual legs of the spiders */
		ComputeRaystoConnect(rayforvertices, numvertices);
		DrawSpider(raystoconnectvtx, numvertices, yellow); 		/* Connects the Vertices (intersection points) */
		mark_cog (rayforspider, numspiderlegs, ltblue);
		secondary_image_redraw ();


}





/*******************************************************************

	Spider main functions:
	pass1 performs the actual edge detection
	pass2 does some post-processing, mainly filtering of r(phi)

*******************************************************************/


void spider_pass1 () 
{
int i, j, k, l;
int prefilter_width;
float angleinrads;
float rotmatrix[2][2];
float initialray[2][2];

	angleinrads = (float) ((2*PI)/numspiderlegs);

	/* The following lets us start at 3 o'clock, counterclockwise */
	
	initialray[0][0] = idiam/2.0;
	initialray[0][1] = 0;
	initialray[1][0] = odiam/2.0;
	initialray[1][1] = 0;
	
	prefilter_width = gauss_prep (prefilter_depth);

	dp (3, "%f %f\n", initialray[0][0], initialray[1][0]);

	for (k = 0; k < numspiderlegs; k++)	  
	{
				
	  	rayforspider[k].angleinrads = (angleinrads*k);	
				    
	  	rayforspider[k].rotmatrix[0][0] = rotmatrix[0][0] = cos(angleinrads*k);
		rayforspider[k].rotmatrix[0][1] = rotmatrix[0][1] = sin(angleinrads*k);
		rayforspider[k].rotmatrix[1][0] = rotmatrix[1][0] = -sin(angleinrads*k);
		rayforspider[k].rotmatrix[1][1] = rotmatrix[1][1] = cos(angleinrads*k);
												
		for (i = 0; i<2; i++)
	  		for (j = 0; j<2; j++)
			{
		    	rayforspider[k].endpoints[i][j] = 0;
		    	for (l=0;l < 2;l++) rayforspider[k].endpoints[i][j] += initialray[i][l]*rotmatrix[l][j];
		    	/* Need to round the values so we can represent pixels  */
		    	rayforspider[k].endpoints[i][j] = floor(0.4999 + (rayforspider[k].endpoints[i][j]));
			}

	    rayforspider[k].startx = (int)rayforspider[k].endpoints[0][0];
	    rayforspider[k].starty = (int)rayforspider[k].endpoints[0][1];
	    rayforspider[k].endx = (int)rayforspider[k].endpoints[1][0];
	    rayforspider[k].endy = (int)rayforspider[k].endpoints[1][1];	

	    /* Now lets calculate the individual pixel coordinates between our endpoints */

   		rayforspider[k].numPixels = BresLine(&rayforspider[k]);

		/* If a prefilter value != 0 is selected, do a Gaussian smoothing */

		if (prefilter_depth > 0.0) gauss_convolve (&rayforspider[k], prefilter_width, /*k==0*/ 0);

		/* Now that we have the spider, let's calculate our cortical intersection points.  
   		 		Flag designates if search goes from inner diameter
				to our outer diameter or vices versa.  0 means inner to outer, 1 means outer to inner */
		FindVertexPoints(&rayforspider[k], spider_dirflg);

	}

				
}			



void spider_pass2 () 
{
		
	use_rejection_threshold();
	filter_data();
	convert_radialtocartesian(rayforvertices, numvertices);
		
}	



/*****************************************************************************

	Spider GUI parts
	
*****************************************************************************/


void spider_statusbar_display (const char* text)
{

guint context;

	context = gtk_statusbar_get_context_id((GtkStatusbar*)spider_statusbar, "spidercontext");

	if (spider_statusbar_was_used) /* pop the last message */
	{
		gtk_statusbar_pop((GtkStatusbar*)spider_statusbar, context);
		spider_statusbar_was_used=0;
	}
	
	gtk_statusbar_push((GtkStatusbar*)spider_statusbar, context, text);
	spider_statusbar_was_used = 1;
}




/* This gets called whenever a control panel element has been changed.
	- read control panel values
	- then start the spider algorithm
*/

void spider_go (GtkWidget* widget, gpointer user_data)
{ 
char s[255];
int level;
	
	if (user_data) level = *(int*)user_data; else level=1;
	idiam = spider_idiam_adj->value;
	odiam = spider_odiam_adj->value;
	if (idiam+5 > odiam) 
	{
		idiam = odiam-5; if (idiam<0) idiam=0;
		spider_idiam_adj->value = idiam;
	}
	
	numspiderlegs = spider_vertex_spin_adj->value;
	rel_thresh = spider_rel_thresh_adj->value;
	prefilter_depth = spider_prefilter_adj->value;
	reject_thresh = spider_reject_adj->value/100.0; 

	numvertices = numspiderlegs;  /* For now this will probably get modified after spider_pass2() */

	/* reload original secondary image window */
	secondary_display (&spider_img, 0, "Spider Op Segmentation");
		
	if ((spider_lastlevel==0) || (level<3))
	{
		if (level==1) spider_pass1();		/* Calculates all the spider elements and vertices based on */
											/* the edge detection method used */
		if ((level==1) || (level==2)) spider_pass2(); 		
				
		spider_draw_results();				/* Draw the spider results into the secondary */
		secondary_image_redraw ();			/* display and use imlib to update display */
		spider_lastlevel=level;
	}

	if ((analysis) && (level==3) && (spider_lastlevel>0))
	{
		if (analysis==1)
			vertebrae_main ( (int)spider_followmax_adj->value, (int)spider_boxsize_adj->value, 
							spider_decay_adj->value/100.0);
		else if (analysis==2)
		{
			spider_draw_results ();
			vessel_meas ((int)spider_followmax_adj->value, (int)spider_boxsize_adj->value, spider_decay_adj->value/100.0);
		}
		spider_lastlevel=level;
	}

	sprintf (s, "%d accepted vertices out of %d", numvertices, numspiderlegs);
	spider_statusbar_display (s);


}


/* The following function invokes a call to the iterate function, where the center of
   gravity of the current polygon is computed, and a new segmentation with this COG as 
   new center is being performed */

void on_spider_iterate_clicked  (GtkButton *button, gpointer user_data)
{
vertex cog;
char s[255];

	if (numvertices==0) return;

	cog = center_of_gravity (rayforspider, numspiderlegs);

	spid_center.x += cog.x;
	spid_center.y += cog.y;

	spider_ofs = spider_create_image (&spider_img, initial_diam, spid_center);
	old_odiam = odiam;
	spider_go (NULL, NULL);
	sprintf (s, "New center: %d %d", spid_center.x, spid_center.y);
	spider_statusbar_display (s);

}


/* Whenever the Z spinbutton is clicked, this function gets called.
   We perform about the same sequence that we'd also perform for iterate */

void spider_zcallback ()
{
vertex cog;

	if (spider_img.data) freebuf (&spider_img);
	spider_ofs = spider_create_image (&spider_img, initial_diam, spid_center);
	old_odiam = odiam;
	spider_go (NULL, NULL);
	
	/* Determine center of gravity, recompute, because the 
	   object may be off-center or placed diagonally */

	if (numvertices>0) 
	{
		cog = center_of_gravity (rayforspider, numspiderlegs);
		
		if ( (abs(cog.x)>1) || (abs(cog.y)>1))
		{
			spid_center.x += cog.x;
			spid_center.y += cog.y;

			spider_ofs = spider_create_image (&spider_img, initial_diam, spid_center);
			old_odiam = odiam;
			spider_go (NULL, NULL);
		}
	}

}


void on_spider_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	zslice_spin_unregister ();
	spider_active=0;
	freebuf (&spider_img);
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_spider_compute_clicked  (GtkButton *button, gpointer user_data)
{

	/* Print measurement results, don't discard spider window */
	
	if (analysis==1)
		vertebrae_meas ((int)spider_followmax_adj->value, (int)spider_boxsize_adj->value, 
							spider_decay_adj->value/100.0);
	else if (analysis==2)
		vessel_meas_all ((int)spider_followmax_adj->value, (int)spider_boxsize_adj->value, spider_decay_adj->value/100.0);


}


/* Spider_ok must transfer the vertices found into the polygonal ROI */


void on_spider_ok_clicked  (GtkButton *button, gpointer user_data)
{
int i;
FILE* dumpfile;
char dumpfilename[1024];
vertex v;

	zslice_spin_unregister ();

	/* Print measurement results */
	
	if (analysis==1)
		vertebrae_meas ((int)spider_followmax_adj->value, (int)spider_boxsize_adj->value, 
							spider_decay_adj->value/100.0);
	else if (analysis==2)
		vessel_meas ((int)spider_followmax_adj->value, (int)spider_boxsize_adj->value, spider_decay_adj->value/100.0);

	/* Copy vertices found into ROI array */
	
	current_roi.roi_mode=ROI_POLY;	/* polygon ROI */
	current_roi.bbox_x1 = current_roi.bbox_x2 = current_roi.centfromspider_x = spid_center.x;
	current_roi.bbox_y1 = current_roi.bbox_y2 = current_roi.centfromspider_y = spid_center.y;
	for (i=0; i<numvertices; i++)
	{
		current_roi.xi[i] = rayforvertices[i].vtxXPixelCoord+spid_center.x;
		current_roi.yi[i] = rayforvertices[i].vtxYPixelCoord+spid_center.y;
		if (current_roi.bbox_x1 > current_roi.xi[i]) current_roi.bbox_x1 = current_roi.xi[i];
		if (current_roi.bbox_x2 < current_roi.xi[i]) current_roi.bbox_x2 = current_roi.xi[i];
		if (current_roi.bbox_y1 > current_roi.yi[i]) current_roi.bbox_y1 = current_roi.yi[i];
		if (current_roi.bbox_y2 < current_roi.yi[i]) current_roi.bbox_y2 = current_roi.yi[i];
	}
	current_roi.numvertices = numvertices;
	current_roi.xi[numvertices]=-1;
	current_roi.yi[numvertices]=-1;

	/* Dump the vertices into a text file 
	
	strcpy (dumpfilename, mainimg_name);
	strcat (dumpfilename, ".vertices.txt");
	dumpfile = fopen (dumpfilename, "wt");
	if (dumpfile)
	{
		v = center_of_gravity (numvertices);
		fprintf (dumpfile, "Center:     ( %4d  %4d )\n", v.x+spid_center.x,v.y+spid_center.y);
		for (i=0; i<numvertices; i++)
		{
			fprintf (dumpfile, "Vertex %3d: ( %4d  %4d )\n", i+1,
			rayforvertices[i].vtxXPixelCoord+spid_center.x,
			rayforvertices[i].vtxYPixelCoord+spid_center.y);
		}
		fclose (dumpfile);
	}
	*/
	
	spider_active=0;
	freebuf (&spider_img);
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void spider_set_edet_menu (GtkWidget *widget, gpointer user_data)
{

	edet_function = *(int*) user_data;
	spider_go (NULL,&menuopt1);
	
}

void spider_set_lvl2_menu (GtkWidget *widget, gpointer user_data)
{

	lvl2filter = *(int*) user_data;
	spider_go (NULL,&menuopt2);
	
}

void on_spider_checkbutton1_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	spider_dirflg = !spider_dirflg;
	if (spider_dirflg)
		gtk_label_set_text (GTK_LABEL(spider_label10), "Out -> In");
	else
		gtk_label_set_text (GTK_LABEL(spider_label10), "In -> Out");
	
	spider_go (NULL, user_data);

}


void create_spider_wnd (int which_analysis)
{
  GtkWidget *spider_window;
  GtkWidget *spider_vbox1;
  GtkWidget *spider_hseparator2;
  GtkWidget *spider_hseparator3;
  GtkWidget *spider_hseparator4;
  GtkWidget *spider_label2;
  GtkWidget *spider_label3;
  GtkWidget *spider_label4;
  GtkWidget *spider_label5;
  GtkWidget *spider_label6;
  GtkWidget *spider_label7;
  GtkWidget *spider_label8;
  GtkWidget *spider_label9;
  GtkWidget *spider_label11;
  GtkWidget *spider_label12;
  GtkWidget *spider_label13;
  GtkWidget *spider_label14;
  GtkWidget *spider_label15;
  GtkWidget *spider_table3;
  GtkWidget *spider_table4;
  GtkWidget *spider_table5;
  GtkWidget *spider_vertex_spin;
  GtkWidget *spider_idiam_hscale;
  GtkWidget *spider_odiam_hscale;
  GtkWidget *spider_relthresh_hscale;
  GtkWidget *spider_prefilter_hscale;
  GtkWidget *spider_edet_menu;
  GtkWidget *spider_edet_menu_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *spider_level2_menu;
  GtkWidget *spider_level2_menu_menu;
  GtkWidget *spider_reject_hscale;
  GtkWidget *spider_followmax_hscale;
  GtkWidget *spider_boxsize_hscale;
  GtkWidget *spider_decay_hscale;
  GtkWidget *spider_hbox1;
  GtkWidget *spider_ok_button;
  GtkWidget *spider_cancel_button;
  GtkWidget *spider_iterate_button;
  GtkWidget *spider_compute_button;
  GtkWidget *spider_checkbutton1;

  spider_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (spider_window), "spider_window", spider_window);
  gtk_window_set_title (GTK_WINDOW (spider_window), "Spider Segmentation");
  spider_vbox1 = vbox_in_window (spider_window, "spider_vbox1");

  spider_table3 = gtk_table_new (7, 2, FALSE);
  gtk_widget_ref (spider_table3);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_table3", spider_table3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_table3);
  gtk_box_pack_start (GTK_BOX (spider_vbox1), spider_table3, FALSE, FALSE, 0);

  spider_label3 = MakeLabel (spider_window, spider_table3, "Vertices:", "spider_label3", 0,0);
  spider_label4 = MakeLabel (spider_window, spider_table3, "Inner diameter:", "spider_label4", 0,1);
  spider_label5 = MakeLabel (spider_window, spider_table3, "Outer diameter:", "spider_label5", 0,2);
  spider_label11= MakeLabel (spider_window, spider_table3, "Denoising:", "spider_label11", 0,3);
  spider_label9 = MakeLabel (spider_window, spider_table3, "Rel sensitivity:", "spider_label9", 0,4);
  spider_label6 = MakeLabel (spider_window, spider_table3, "Edge detection:", "spider_label6", 0,5);
  spider_label10= MakeLabel (spider_window, spider_table3, "In -> Out", "spider_label10", 0,6);
  
  spider_vertex_spin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (64, 16, MAXVERTICES-1, 1, 10, 10));
  spider_vertex_spin = MakeSpinbutton(spider_vertex_spin_adj,spider_window, spider_table3,
  									"spider_vertex_spin", 1,0);
  
  spider_idiam_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, 0, 1024, 1, 10, 10));
  spider_idiam_hscale = MakeHScale (spider_idiam_adj, spider_window, spider_table3,
  									"spider_idiam_hscale", 1,1);

  spider_odiam_adj = GTK_ADJUSTMENT(gtk_adjustment_new (initial_diam, 20, 1024, 1, 10, 10));
  spider_odiam_hscale = MakeHScale (spider_odiam_adj, spider_window, spider_table3,
  									"spider_odiam_hscale", 1,2);

  spider_prefilter_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, 0, 10, 0.1, 1, 1));
  spider_prefilter_hscale = MakeHScale (spider_prefilter_adj, spider_window, spider_table3,
  									"spider_odiam_hscale", 1,3);

  spider_rel_thresh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (20, 1, 100, 1, 10, 10));
  spider_relthresh_hscale = MakeHScale (spider_rel_thresh_adj, spider_window, spider_table3,
  									"spider_odiam_hscale", 1,4);
  
  spider_edet_menu = Make_Optionmenu (spider_window,spider_table3,"spider_edet_menu",1,5);
  gtk_container_set_border_width (GTK_CONTAINER (spider_edet_menu), 1);
  spider_edet_menu_menu = gtk_menu_new ();
  glade_menuitem = Option_menuitem (spider_edet_menu_menu, GTK_SIGNAL_FUNC(spider_set_edet_menu), "Gradient", (gpointer) &menuopt0);
  glade_menuitem = Option_menuitem (spider_edet_menu_menu, GTK_SIGNAL_FUNC(spider_set_edet_menu), "Maximum", (gpointer) &menuopt1);
  glade_menuitem = Option_menuitem (spider_edet_menu_menu, GTK_SIGNAL_FUNC(spider_set_edet_menu), "Minimum", (gpointer) &menuopt2);
  glade_menuitem = Option_menuitem (spider_edet_menu_menu, GTK_SIGNAL_FUNC(spider_set_edet_menu), "Relative", (gpointer) &menuopt3);
  glade_menuitem = Option_menuitem (spider_edet_menu_menu, GTK_SIGNAL_FUNC(spider_set_edet_menu), "Local Variance", (gpointer) &menuopt4);

  gtk_option_menu_set_menu (GTK_OPTION_MENU (spider_edet_menu), spider_edet_menu_menu);
  gtk_option_menu_set_history (GTK_OPTION_MENU (spider_edet_menu), 0);

  spider_checkbutton1 = gtk_check_button_new_with_label ("Search direction");
  gtk_widget_ref (spider_checkbutton1);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_checkbutton1", spider_checkbutton1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_checkbutton1);
  gtk_table_attach (GTK_TABLE (spider_table3), spider_checkbutton1, 1, 2, 6, 7,
                    (GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);

  gtk_signal_connect (GTK_OBJECT (spider_vertex_spin_adj), "value_changed",
                      GTK_SIGNAL_FUNC (spider_go), &menuopt1);
  gtk_signal_connect (GTK_OBJECT (spider_idiam_adj), "value_changed",
                      GTK_SIGNAL_FUNC (spider_go), &menuopt1);
  gtk_signal_connect (GTK_OBJECT (spider_odiam_adj), "value_changed",
                      GTK_SIGNAL_FUNC (spider_go), &menuopt1);
  gtk_signal_connect (GTK_OBJECT (spider_rel_thresh_adj), "value_changed",
                      GTK_SIGNAL_FUNC (spider_go), &menuopt1);
  gtk_signal_connect (GTK_OBJECT (spider_prefilter_adj), "value_changed",
                      GTK_SIGNAL_FUNC (spider_go), &menuopt1);


  /* Level 2 starting with hseparator and "Level 2" label */

  spider_hseparator2 = MakeHsep (spider_window,spider_vbox1,"spider_hseparator2",-1,-1);
  spider_label2 = MakeLabel (spider_window, spider_vbox1, "Level 2 parameters", "spider_label2", -1,-1);

  spider_table4 = gtk_table_new (2, 2, FALSE);
  gtk_widget_ref (spider_table4);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_table4", spider_table4,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_table4);
  gtk_box_pack_start (GTK_BOX (spider_vbox1), spider_table4, FALSE, FALSE, 0);

  spider_label7 = MakeLabel (spider_window, spider_table4, "Filter:", "spider_label7", 0,0);
  spider_label8 = MakeLabel (spider_window, spider_table4, "Rejection threshold:", "spider_label8", 0,1);

  spider_level2_menu = Make_Optionmenu (spider_window,spider_table4,"spider_level2_menu",1,0);
  spider_level2_menu_menu = gtk_menu_new ();
  glade_menuitem = Option_menuitem (spider_level2_menu_menu, GTK_SIGNAL_FUNC(spider_set_lvl2_menu), "None", (gpointer) &menuopt0);
  glade_menuitem = Option_menuitem (spider_level2_menu_menu, GTK_SIGNAL_FUNC(spider_set_lvl2_menu), "Weak", (gpointer) &menuopt1);
  glade_menuitem = Option_menuitem (spider_level2_menu_menu, GTK_SIGNAL_FUNC(spider_set_lvl2_menu), "Medium", (gpointer) &menuopt2);
  glade_menuitem = Option_menuitem (spider_level2_menu_menu, GTK_SIGNAL_FUNC(spider_set_lvl2_menu), "Agressive", (gpointer) &menuopt3);

  gtk_option_menu_set_menu (GTK_OPTION_MENU (spider_level2_menu), spider_level2_menu_menu);
  gtk_option_menu_set_history (GTK_OPTION_MENU (spider_level2_menu), 0);

  spider_reject_adj = GTK_ADJUSTMENT(gtk_adjustment_new (50, 0, 100, 1, 10, 10));
  spider_reject_hscale = MakeHScale (spider_reject_adj, spider_window, spider_table4,
  									"spider_reject_hscale", 1,1);
  
  gtk_signal_connect (GTK_OBJECT (spider_reject_adj), "value_changed",
                      GTK_SIGNAL_FUNC (spider_go), &menuopt2);
  gtk_signal_connect (GTK_OBJECT (spider_checkbutton1), "toggled",
                      GTK_SIGNAL_FUNC (on_spider_checkbutton1_toggled), &menuopt1);

  /* Level 3 starting with hseparator. Level 3 does not appear in the general spider op. */

	if (which_analysis)
	{
		spider_hseparator3 = MakeHsep (spider_window,spider_vbox1,"spider_hseparator3",-1,-1);
		spider_label12 = MakeLabel (spider_window, spider_vbox1, "Level 3 parameters", "spider_labe12", -1,-1);
		gtk_misc_set_padding (GTK_MISC (spider_label12), 0, 5);
		
		spider_table5 = gtk_table_new (3, 2, FALSE);
		gtk_widget_ref (spider_table5);
		gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_table5", spider_table5,
									(GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (spider_table5);
		gtk_box_pack_start (GTK_BOX (spider_vbox1), spider_table5, FALSE, FALSE, 0);
		
		if (which_analysis==1)
		{
			spider_label13 = MakeLabel (spider_window, spider_table5, "Follow max", "spider_label13", 0,0);
			spider_label14 = MakeLabel (spider_window, spider_table5, "Max thickness", "spider_label14", 0,1);
			spider_label15 = MakeLabel (spider_window, spider_table5, "Cortex decay", "spider_label15", 0,2);
		}
		else
		{
			spider_label13 = MakeLabel (spider_window, spider_table5, "Search length", "spider_label13", 0,0);
			spider_label14 = MakeLabel (spider_window, spider_table5, "Max thickness", "spider_label14", 0,1);
			spider_label15 = MakeLabel (spider_window, spider_table5, "Threshold", "spider_label15", 0,2);
		}

		spider_followmax_adj = GTK_ADJUSTMENT(gtk_adjustment_new (2, 0, 10, 1, 1, 1));
		spider_followmax_hscale = MakeHScale (spider_followmax_adj, spider_window, spider_table5,
											"spider_followmax_hscale", 1,0);
		
		spider_boxsize_adj = GTK_ADJUSTMENT(gtk_adjustment_new (10, 0, 20, 1, 1, 1));
		spider_boxsize_hscale = MakeHScale (spider_boxsize_adj, spider_window, spider_table5,
											"spider_boxsize_hscale", 1,1);
		
		spider_decay_adj = GTK_ADJUSTMENT(gtk_adjustment_new (60, 0, 100, 1, 10, 1));
		spider_decay_hscale = MakeHScale (spider_decay_adj, spider_window, spider_table5,
											"spider_decay_hscale", 1,2);

		gtk_signal_connect (GTK_OBJECT (spider_followmax_adj), "value_changed",
							GTK_SIGNAL_FUNC (spider_go), &menuopt3);
		gtk_signal_connect (GTK_OBJECT (spider_boxsize_adj), "value_changed",
							GTK_SIGNAL_FUNC (spider_go), &menuopt3);
		gtk_signal_connect (GTK_OBJECT (spider_decay_adj), "value_changed",
							GTK_SIGNAL_FUNC (spider_go), &menuopt3);
	}


  /* H-box to contain the three buttons */


  spider_hseparator4 = MakeHsep (spider_window,spider_vbox1,"spider_hseparator4",-1,-1);

  spider_hbox1 = gtk_hbox_new (TRUE, 0);
  gtk_widget_ref (spider_hbox1);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_hbox1", spider_hbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_hbox1);
  gtk_box_pack_start (GTK_BOX (spider_vbox1), spider_hbox1, TRUE, TRUE, 0);

  spider_ok_button = gtk_button_new_with_label ("OK");
  gtk_widget_ref (spider_ok_button);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_ok_button", spider_ok_button,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_ok_button);
  gtk_box_pack_start (GTK_BOX (spider_hbox1), spider_ok_button, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (spider_ok_button), 2);

  spider_cancel_button = gtk_button_new_with_label ("Cancel");
  gtk_widget_ref (spider_cancel_button);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_cancel_button", spider_cancel_button,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_cancel_button);
  gtk_box_pack_start (GTK_BOX (spider_hbox1), spider_cancel_button, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (spider_cancel_button), 2);

  spider_iterate_button = gtk_button_new_with_label ("Iterate");
  gtk_widget_ref (spider_iterate_button);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_iterate_button", spider_iterate_button,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_iterate_button);
  gtk_box_pack_start (GTK_BOX (spider_hbox1), spider_iterate_button, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (spider_iterate_button), 2);

	if (which_analysis)
	{

		spider_compute_button = gtk_button_new_with_label ("Compute");
		gtk_widget_ref (spider_compute_button);
		gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_compute_button", spider_compute_button,
									(GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show (spider_compute_button);
		gtk_box_pack_start (GTK_BOX (spider_hbox1), spider_compute_button, TRUE, TRUE, 0);
		gtk_container_set_border_width (GTK_CONTAINER (spider_compute_button), 2);
		gtk_signal_connect (GTK_OBJECT (spider_compute_button), "clicked",
                      GTK_SIGNAL_FUNC (on_spider_compute_clicked), spider_window);
	}

  spider_statusbar = gtk_statusbar_new ();
  gtk_widget_ref (spider_statusbar);
  gtk_object_set_data_full (GTK_OBJECT (spider_window), "spider_statusbar", spider_statusbar,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (spider_statusbar);
  gtk_box_pack_start (GTK_BOX (spider_vbox1), spider_statusbar, FALSE, FALSE, 0);

  /* Buttons and the destroy signal */

  gtk_signal_connect (GTK_OBJECT (spider_ok_button), "clicked",
                      GTK_SIGNAL_FUNC (on_spider_ok_clicked), spider_window);
  gtk_signal_connect (GTK_OBJECT (spider_cancel_button), "clicked",
                      GTK_SIGNAL_FUNC (on_spider_cancel_clicked), spider_window);
  gtk_signal_connect (GTK_OBJECT (spider_iterate_button), "clicked",
                      GTK_SIGNAL_FUNC (on_spider_iterate_clicked), spider_window);
  gtk_signal_connect (GTK_OBJECT (spider_window), "destroy",
                      GTK_SIGNAL_FUNC (on_spider_cancel_clicked), spider_window);


  gtk_widget_show (spider_window);
}





/*************************************************

	Spider Op: Main
	There is a general spider op (which_analysis=0) without special functions,
	a spider op for vertebral bone analysis (which_analysis=1), and
	a spider op for blood vessel analysis in CT or MR (which_analysis=2)
	
**************************************************/



void spider_main (int which_analysis)
{

	if (snake_active)
	{
		ShowMessage ("Snake and Spider segmentation are mutually exclusive");
		return;
	}
	
	
	analysis = which_analysis;
	if (clickstack_last<1)
	{
		ShowMessage ("You must indicate (click) center of ROI first");
		return;
	}

	spid_center.x = clickstack[clickstack_last-1].x;
	spid_center.y = clickstack[clickstack_last-1].y;

	/* Clone and crop a section of the original image centered
	   on the last clicked coordinates, then display it in
	   the secondary window */
	
	initial_diam = INITIAL_ODIAM;
	spider_img.data = NULL;
	if (initial_diam > MIN (mainimg.xmax,mainimg.ymax)) initial_diam = MIN (mainimg.xmax,mainimg.ymax);
	spider_ofs = spider_create_image (&spider_img, (12*initial_diam)/10, spid_center);
	secondary_display (&spider_img, 0, "Spider Op Segmentation");

	spider_statusbar_was_used=0;
	spider_dirflg=0;
	numvertices=0;
	spider_lastlevel=0;
	old_odiam = INITIAL_ODIAM;
	spider_active=1;
	
	zslice_spin_register ((gpointer)spider_zcallback);
	
	create_spider_wnd (which_analysis);
	

}



